package il.ac.tau.cs.dqsq.axmlservices;

import javax.xml.parsers.DocumentBuilderFactory;

//import org.w3c.dom.Document;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.io.File;
import fr.inria.gemo.util.xml.DOMSerializer;
import il.ac.tau.cs.dqsq.util.dQSQUtils;
import il.ac.tau.cs.dqsq.termdetection.ActivityService;


/**
 * @author pettelno
 *
 * The DiffService class exposes the doDiff web service which is activated
 * on each reply of a continuous service before being sent to the client.
 * The parameter contServiceReply is the reply that was generated by
 * the continuous service. The doDiff service is responsible for sending only
 * new data (i.e., the delta). This goal is achieved by keeping the history of
 * the data sent to each client in the ContService_X_clients.xml file and
 * comparing every reply to this history.
 *
 * The doDiff service also takes part in the termination detection algorithm:
 * After checking new data for a client, decrement activity if there is no
 * new data. 
 * 
 * Note: The scope of the DiffService class, as defined in server-config.wsdd
 * is "application", which means that one object is used for the whole run.
 *   
 */
public class DiffService
{
	private boolean isTerminationDetectionOn;
	private String repositoryLocation = null;
	
	public DiffService()
	{
		this.isTerminationDetectionOn = dQSQUtils.isTerminationDetectionOn();
		this.repositoryLocation = dQSQUtils.getRepositoryLocation();
	}
	
	public Element doDiff(Element contServiceReply)
	{
		String serviceName = null;
		String clientId = null;
		
		try
		{
			if(contServiceReply == null)
			{
				dQSQUtils.debug("ERROR - DiffService - Diff service received null reply.", "debug_DiffService.txt");
				return null;
			}
			
			dQSQUtils.debug("DEBUG - DiffService received:\n" + DOMSerializer.Serialize(contServiceReply), "debug_DiffService.txt");

			Node serviceNameNode = contServiceReply.getElementsByTagName("serviceName").item(0); 
			serviceName = serviceNameNode.getFirstChild().getNodeValue();
			Node servicePeerNode = contServiceReply.getElementsByTagName("servicePeer").item(0); 
			String servicePeer = servicePeerNode.getFirstChild().getNodeValue();
	
			// If the peer is terminated (termination was detected) - 
			// Return to the client a call to this service with
			// destroy="true", to terminate the service.
			if(new ActivityService().isTerminated(servicePeer))
				return destroyService(serviceName, servicePeer);
			
			String clientsFileName = dQSQUtils.getRepositoryLocation() + File.separator + serviceName + "_clients.xml";
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();

			Node clientIdNode = contServiceReply.getElementsByTagName("clientID").item(0); 
			clientId = clientIdNode.getFirstChild().getNodeValue();

			String clientRelationName = null;
			if(!dQSQUtils.isLocalRule(serviceName))
			{
				Node clientRelationNameNode = contServiceReply.getElementsByTagName("clientRelationName").item(0); 
				clientRelationName = clientRelationNameNode.getFirstChild().getNodeValue();
			}
			
			boolean areNewTuples = false;
			boolean origReplyEmpty = true;
			synchronized(RegisterService.clientsFileSync)
			{
				Document clientsDoc = dbf.newDocumentBuilder().parse(new File(clientsFileName));
				Element docElem = clientsDoc.getDocumentElement();
	
				NodeList clientElems = docElem.getElementsByTagName("client");
				int clientsNum = clientElems.getLength();
				
				// Get the client node corresponding to the client ID from the
				// clients document of the service.
				Element clientElem = null;
				for(int i = 0; i < clientsNum; i++)
				{
					Element curClientElem = (Element)clientElems.item(i);
					String curClientId = curClientElem.getAttribute("id");
					
					if(curClientId.equals(clientId))
					{
						clientElem = curClientElem;
						break;
					}
				}
				if(clientElem == null)
				{
					dQSQUtils.debug("ERROR - DiffService: Client with ID " + clientId + " did not register on service " + serviceName, "debug_DiffService.txt");
					return null; // client did not register on the continuous service
				}
				
				// clientNode is now the node corresponding to the client.
				
				// If the "shouldCheckData" attribute is set to false,
				// there is no new date for this client.
				// The value of the "shouldCheckData" attribute is true only when
				// the service potentially has new data for the client. This is
				// true in 2 cases:
				// 1) Immediately after the registration.
				// 2) Immediately after the service receives new data from an
				//    external resource.

				
				clientElem.setAttribute("shouldCheckData", "false");
				dQSQUtils.debug("DEBUG - DiffService: Setting the shouldCheckData attribute to false for " + serviceName, "debug_DiffService.txt");
				
				Node clientTuplesNode = clientElem.getFirstChild();
				Node replyTuplesNode = contServiceReply.getElementsByTagName("tuples").item(0);
				
				// Compare the reply before diff to what already sent to this
				// client and update accordingly the reply and the clients document.
				// The comparison is currently not implemented efficiently,
				// as every tuple in the reply is searched in the clients document
				// i.e., O(n^2).
				if(replyTuplesNode.getChildNodes().getLength() > 0) // if length is 0, it's an empty reply
				{
					origReplyEmpty = false;
					Node curReplyTuple = replyTuplesNode.getFirstChild();
					while(curReplyTuple != null)
					{
						if(isNewTuple(curReplyTuple, clientTuplesNode))
						{
							if(clientTuplesNode == null)
							{
								clientTuplesNode = clientsDoc.createElement("tuples");
								clientElem.appendChild(clientTuplesNode);
							}
								
							addTupleToClientsDoc(curReplyTuple, clientTuplesNode, clientsDoc);
							areNewTuples = true;
							curReplyTuple = curReplyTuple.getNextSibling();
						}
						else
						{
							Node tupleToRemove = curReplyTuple;
							curReplyTuple = curReplyTuple.getNextSibling();
							replyTuplesNode.removeChild(tupleToRemove);
						}
					}
				}
			
				// We need to flush the clients document even if there are
				// no new tuples to send, because the value of the
				// "shouldCheckData" was set to "false". 
				dQSQUtils.flushDocument(clientsDoc, clientsFileName);	
			}
			
			if(areNewTuples)
			{
				dQSQUtils.debug("DEBUG - DiffService: Reply for service " + serviceName + ":\n " + DOMSerializer.Serialize(contServiceReply), "debug_DiffService.txt");
				dQSQUtils.debug("DEBUG - DiffService: Service " + serviceName + " sent new data for client with ID " + clientId, "debug_DiffService.txt");

				if(!dQSQUtils.isLocalRule(serviceName))
					dQSQUtils.debug("DEBUG - " + serviceName + " sent new data for client " + clientRelationName + " with ID " + clientId, "debug_Activity.txt");

				return contServiceReply;
			}
			else
			{
				// The service checked for new data for the client,
				// which means that the service potentially had new
				// data for for the client.
				// There is no new data, so decrement activity of the
				// service's node (for the termination detection algorithm) -
				// only if it's a service and not a local rule.
				if(!dQSQUtils.isLocalRule(serviceName))
					new ActivityService().decrementActivity(dQSQUtils.service2Relation(serviceName), "rule 7, after checking new data with no reply for client " + clientRelationName, servicePeer);
					
				// No new tuples - don't wake up client.
				dQSQUtils.debug("DEBUG - DiffService: Reply for service " + serviceName + ":\n " + DOMSerializer.Serialize(contServiceReply), "debug_DiffService.txt");
				dQSQUtils.debug("DEBUG - DiffService: No new data for client " + clientId + " of service " + serviceName, "debug_DiffService.txt");
								
				return null;
			}
		}
		catch(Exception ex)
		{
			dQSQUtils.debug("ERROR - DiffService - Diff service failed. Service: " + serviceName + ". Client ID: " + clientId + "System error: " + ex.getMessage(), "debug_DiffService.txt");
			return null;
		}
	}
	
	private void addTupleToClientsDoc(Node tupleNodeToAdd, Node clientTuplesNode, Document clientsDoc)
	{
		Node importedReplyTupleNode = clientsDoc.importNode(tupleNodeToAdd, true);
		clientTuplesNode.appendChild(importedReplyTupleNode);		
	}

	private boolean isNewTuple(Node curReplyTuple, Node clientTuplesNode)
	{
		if(clientTuplesNode == null)
			return true; // no tuples in client node.
		
		Node curClientTuple = clientTuplesNode.getFirstChild();
		while(curClientTuple != null)
		{
			if(dQSQUtils.areNodesEqual(curReplyTuple, curClientTuple))
				return false;
			
			curClientTuple = curClientTuple.getNextSibling();
		}
		
		// Tuple does not exist in client node.
		return true;
	}
	
	private Element destroyService(String serviceName, String servicePeer)
	{
		// For local rules (e.g., "Rule_r1") the serviceName element
		// in the reply is "Rule_r1". Change it to "ContRule_r1",
		// in order to destroy it.
		if(dQSQUtils.isLocalRule(serviceName))
			serviceName = "Cont" + serviceName;
		
		dQSQUtils.debug("DEBUG - Destroying service " + serviceName, "debug_DiffService.txt");

		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		
		Document tmpDoc = null;
		try
		{
			tmpDoc = dbf.newDocumentBuilder().newDocument();
		}
		catch(Exception ex)
		{
			dQSQUtils.debug("ERROR - DiffService failed to create new document to destroy service.", "debug_DiffService.txt");
			return null; // exception
		}
		
		Element destroyServiceElem = tmpDoc.createElementNS("http://www-rocq.inria.fr/verso/AXML", "sc");
		destroyServiceElem.setAttributeNS("http://www.w3.org/2000/xmlns/",
								   "xmlns:axml", 
								   "http://www-rocq.inria.fr/verso/AXML");
		destroyServiceElem.setAttribute("frequency", "once");
		destroyServiceElem.setAttribute("methodName", serviceName);
		destroyServiceElem.setAttribute("serviceNameSpace", serviceName);
		destroyServiceElem.setAttribute("serviceURL", dQSQUtils.getPeerURL() + "/servlet/AxisServlet");
		destroyServiceElem.setAttribute("subscription", "true");
		destroyServiceElem.setAttribute("destroy", "true");
		
		return destroyServiceElem;
	}	
}
