package il.ac.tau.cs.dqsq.axmlservices;

import java.net.URL;
import localhost.axml.services.IdlenessManager.IdlenessManagerServiceLocator;
import localhost.axml.services.PeerAccessPoint.PeerAccessPointServiceLocator;
import localhost.axml.services.AckService.AckServiceServiceLocator;
import il.ac.tau.cs.dqsq.util.Logger;
import il.ac.tau.cs.dqsq.util.dQSQUtils;
import il.ac.tau.cs.dqsq.termdetection.ActivityService;

/**
 * @author pettelno
 *
 * The purpose of this class is to expose web services that gather a few calls
 * to dQSQ control services. These services are called from the help documents.
 * By gathering a few control services we gain a single call from the AXML
 * document instead of multiple calls with "followedBy" between them.
 * Note that currently this class is used only for the reply that is returned
 * from the last relation in a rule. (Originally there were 4 services returned
 * to the help document. They were replaced with a single call to 
 * postRecvDataFromLastService). Among the services returned from the first
 * relation in a rule are services that must be physically located in the help
 * document, therefore GeneralService is not used in this case.
 * 
 */
public class GeneralService
{
	/**
	 * This method is exposed as a web service. It is called after the 
	 * relation in the head of a rule receives data from the last relation
	 * in the rule body. 
	 * 
	 * @param serviceName The name of the service that sent the data.
	 * @param clientRelationName The name of the relation that received
	 *        the data.
	 * @param clientURL The URL of the relation that received the data (client).
	 * @param serverURL The URL of the service that sent the data (server).
	 * @param clientPeer The name of the client's peer. 
	 * @param serverPeer The name of the server's peer. 
	 * 
	 */
	public void postRecvDataFromLastService(String serviceName,
			String clientRelationName, String clientURL, String serverURL,
			String clientPeer, String serverPeer)
	{
		// Termination detection logic after client receives new data from last
		// relation in rule (second relation if there are 2 relations in rule
		// body, or only relation if there is a single relation in rule body).

		synchronized(RegisterService.clientsFileSync)
		{			
			// 1. Increment local activity value.
			new ActivityService().incrementActivityAfterDataFromSecondService(clientRelationName, clientPeer, serviceName);

			// 2. Call the onServiceEvent web service of the client relation's
			// service. For example, if R receives data from P, then onServiceEvent
			// is called for ContService_R so that R will check for new data
			// for his own clients.
			PeerAccessPointServiceLocator locator = new PeerAccessPointServiceLocator();
			try
			{
				URL serviceURL = new URL(clientURL + "/services/PeerAccessPoint");
				localhost.axml.services.PeerAccessPoint.PeerAccessPoint service = locator.getPeerAccessPoint(serviceURL);
				service.onServiceEvent(dQSQUtils.relation2Service(clientRelationName));
			}
			catch(Exception e)
			{
				Logger.sendMessage("Failed to invoke the onServiceEvent web service for service " + dQSQUtils.relation2Service(clientRelationName) + ".",
								   Logger.LOG_LEVEL_ERROR);        	
			}			
			
			// 3. Call local IdlenessManager.onRecvMsg.
			IdlenessManagerServiceLocator imLocator = new IdlenessManagerServiceLocator();
			try
			{
				URL serviceURL = new URL(clientURL + "/services/IdlenessManager");
				localhost.axml.services.IdlenessManager.IdlenessManager service = imLocator.getIdlenessManager(serviceURL);
				service.onRecvMsg(clientRelationName);
			}
			catch(Exception e)
			{
				dQSQUtils.debug("ERROR - AXMLGate failed to call IdlenessManager.onRecvMsg for relation " + clientRelationName + ". System error: " + e.getMessage(), "debug_AXMLGate.txt");
			}

			// 4. Send ack on data to remote service.
			AckServiceServiceLocator ackLocator = new AckServiceServiceLocator();
			try {
				URL ackServiceURL = new URL(serverURL + "/services/AckService");
				localhost.axml.services.AckService.AckService ackService = ackLocator.getAckService(ackServiceURL);
				ackService.recvAck(serviceName, clientRelationName, serverPeer, "after receiving data from second relation");
			}
			catch(Exception e)
			{
				dQSQUtils.debug("ERROR - Relation " + clientRelationName + " failed to send ack to service " + serviceName + " after receiving data. System error: " + e.getMessage(), "debug_AXMLGate.txt");			
			}

			// 5. Decrement local activity value.
			new ActivityService().decrementActivity(clientRelationName, "rule 6, after sending ack on data to " + serviceName, clientPeer);
				
			// 6. Set the "shouldCheckData" attribute of the service of the 
			// relation that received the data to "true".
			new RegisterService().setShouldCheckData(clientRelationName, clientPeer);
		}		
	}
	
	public void onReceiveData(String receivingRelationName)
	{
		// Call the onServiceEvent web service of the client relation's
		// service. For example, if R receives data from P, then onServiceEvent
		// is called for ContService_R so that R will check for new data
		// for his own clients.
		PeerAccessPointServiceLocator locator = new PeerAccessPointServiceLocator();
		try
		{
			URL serviceURL = new URL(dQSQUtils.getPeerURL() + "/services/PeerAccessPoint");
			localhost.axml.services.PeerAccessPoint.PeerAccessPoint service = locator.getPeerAccessPoint(serviceURL);
			service.onServiceEvent("Cont" + receivingRelationName);
		}
		catch(Exception e)
		{
			Logger.sendMessage("Failed to invoke the onServiceEvent web service for service " + "Cont" + receivingRelationName + ".",
							   Logger.LOG_LEVEL_ERROR);        	
		}					
	}
}
