package il.ac.tau.cs.dqsq.axmlservices;

import java.io.File;

import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import localhost.axml.services.IdlenessManager.IdlenessManagerServiceLocator;
import localhost.axml.services.AckService.AckServiceServiceLocator;
import java.net.URL;
import il.ac.tau.cs.dqsq.util.dQSQUtils;
import il.ac.tau.cs.dqsq.termdetection.ActivityService;

/**
 * @author pettelno
 *
 * Provides web services that manage registration of clients on continuous
 * AXML services.
 * 
 * Note: The scope of the RegisterService class, as defined in server-config.wsdd,
 * is "application", which means that one object is used for the whole run.
 */
public class RegisterService
{
	static String clientsFileSync = "clientsFileSync";
	private String repositoryLocation = null;

	public RegisterService()
	{
		this.repositoryLocation = dQSQUtils.getRepositoryLocation();
	}
	
	/**
	 * This method is exposed as a web service. Before every service call
	 * to a continuous AXML service, this service is called to register
	 * the client on the service. For every continuous service ContService_X
	 * there is a corresponding file ContService_X_clients.xml that keeps
	 * track of all the clients of that service and the accumulated data
	 * that was sent to every client. The history of sent data is kept to
	 * prevent the service from sending duplicated data to a client. 
	 * 
	 * @param serviceName The name of the service to register on.
	 * @param clientID A unique ID that identifies the registering client.
	 * @param clientURL The URL of the client.
	 * @param clientRelationName The relation name of the client.
	 * @param servicePeer The service peer's name.
	 * @param clientPeer The client peer's name.
	 * 
	 */
	public void register(String serviceName, String clientID, String clientURL, String clientRelationName, String servicePeer, String clientPeer)
	{
		// If this is a service (e.g., "ContService_sup2")
		// and not a local rule (e.g., "ContRule_r1"), we:
		// 1. Increment activity value.
		// 2. Update idleness value.
		if(!dQSQUtils.isLocalRule(serviceName))
		{
			new ActivityService().incrementActivity(dQSQUtils.service2Relation(serviceName), "rule 3, after receiving register request from " + clientRelationName, servicePeer);
			updateIdleness(serviceName, servicePeer);
		}
		
		synchronized(clientsFileSync)
		{
			String clientsFileName = this.repositoryLocation + File.separator + serviceName + "_clients.xml";
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
	
			Document clientsDoc = null; 
			try
			{
				clientsDoc = dbf.newDocumentBuilder().parse(new File(clientsFileName));
			}
			catch(Exception ex)
			{
				dQSQUtils.debug("ERROR - RegisterService failed to parse clients file - " + clientsFileName, "debug_RegisterService.txt");
				return;
			}
				
			Element docElem = clientsDoc.getDocumentElement();
		
			if(isClientRegistered(clientID, docElem))
			{
				dQSQUtils.debug("WARNING - Client " + clientRelationName + " already registered on service " + serviceName, "debug_RegisterService.txt");			
				return;
			}
		
			dQSQUtils.debug("DEBUG - Client " + clientRelationName + " (URL: " + clientURL + ") registered on service " + serviceName, "debug_RegisterService.txt");			
						
			Element clientElem = clientsDoc.createElement("client");
			clientElem.setAttribute("id", clientID);
			clientElem.setAttribute("shouldCheckData", "true");
			docElem.appendChild(clientElem);
			dQSQUtils.flushDocument(clientsDoc, clientsFileName);
		}

		// Send ack.
		if(clientURL != null && clientURL.length() > 0)
		{
			AckServiceServiceLocator locator = new AckServiceServiceLocator();
			try
			{
				URL ackServiceURL = new URL(clientURL + "/services/AckService");
				localhost.axml.services.AckService.AckService ackService = locator.getAckService(ackServiceURL);
				ackService.recvAck(clientRelationName, serviceName, clientPeer, "after registering on service");
			}
			catch(Exception e)
			{
				dQSQUtils.debug("ERROR - Service " + serviceName + " failed to send ack to client " + clientRelationName + " (URL: " + clientURL + ") after registering. System error: " + e.getMessage(), "debug_RegisterService.txt");			
			}
		}				
	}

	/**
	 * This method is exposed as a web service. It sets the "shouldCheckData"
	 * attribute of the client element to "true" for all the registered
	 * clients. After relation X receives new data, it calls this service.
	 * 
	 */
	public void setShouldCheckData(String relationName, String peer) 
	{
		dQSQUtils.debug("DEBUG - Setting the shouldCheckData attribute to true for " + relationName, "debug_RegisterService.txt");		

		String clientsFileName = null;
		if(dQSQUtils.isLocalRule(relationName))
			clientsFileName = relationName + "_clients.xml";
		else
			clientsFileName = dQSQUtils.relation2Service(relationName) + "_clients.xml";
			
		String clientsFilePath = this.repositoryLocation + File.separator + clientsFileName;
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		int numOfPassiveClients = 0;

		Document clientsDoc = null; 
//		synchronized(clientsFileSync)
//		{
			try
			{
				clientsDoc = dbf.newDocumentBuilder().parse(new File(clientsFilePath));
			}
			catch(Exception ex)
			{
				dQSQUtils.debug("ERROR - RegisterService failed to parse clients file - " + clientsFilePath, "debug_RegisterService.txt");
				return;
			}
				
			Element docElem = clientsDoc.getDocumentElement();
		
			NodeList clientElems = docElem.getElementsByTagName("client");
			int clientsNum = clientElems.getLength();
				
			for(int i = 0; i < clientsNum; i++)
			{
				Element curClientElement = (Element)clientElems.item(i);
				if(curClientElement.getAttribute("shouldCheckData").equals("false"))
				{
					curClientElement.setAttribute("shouldCheckData", "true");
					numOfPassiveClients++;
				}
			}
				
			if(numOfPassiveClients > 0)
				dQSQUtils.flushDocument(clientsDoc, clientsFilePath);
//		}
	}

	public int getNumOfPassiveClients(String relationName, String peer) throws Exception
	{
		// Returns the number of registered clients for which the current
		// value of the "shouldCheckData" attribute is "false". We say on
		// such a client that it is passive.
		
		String clientsFileName = this.repositoryLocation + File.separator + dQSQUtils.relation2Service(relationName) + "_clients.xml";
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		int numOfPassiveClients = 0;

		Document clientsDoc = null; 
		try
		{
			clientsDoc = dbf.newDocumentBuilder().parse(new File(clientsFileName));
		}
		catch(Exception ex)
		{
			dQSQUtils.debug("ERROR - RegisterService failed to parse clients file - " + clientsFileName, "debug_RegisterService.txt");
			throw ex;
		}
			
		Element docElem = clientsDoc.getDocumentElement();
	
		NodeList clientElems = docElem.getElementsByTagName("client");
		int clientsNum = clientElems.getLength();
		dQSQUtils.debug("DEBUG - " + relationName + " has " +  clientsNum + " clients", "debug_RegisterService.txt");
			
		for(int i = 0; i < clientsNum; i++)
		{
			Element curClientElement = (Element)clientElems.item(i);
			if(curClientElement.getAttribute("shouldCheckData").equals("false"))
			{
				numOfPassiveClients++;
				dQSQUtils.debug("DEBUG - client " + i + " of " + relationName + " is passive", "debug_RegisterService.txt");
			}
			else
				dQSQUtils.debug("DEBUG - client " + i + " of " + relationName + " is active", "debug_RegisterService.txt");
		}
		
		return numOfPassiveClients;
	}

	public int getNumOfClients(String relationName, String peer) throws Exception
	{
		String clientsFileName = this.repositoryLocation + File.separator + dQSQUtils.relation2Service(relationName) + "_clients.xml";
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		int numOfPassiveClients = 0;

		Document clientsDoc = null; 
		try
		{
			clientsDoc = dbf.newDocumentBuilder().parse(new File(clientsFileName));
		}
		catch(Exception ex)
		{
			dQSQUtils.debug("ERROR - RegisterService failed to parse clients file - " + clientsFileName, "debug_RegisterService.txt");
			throw ex;
		}
			
		Element docElem = clientsDoc.getDocumentElement();
	
		NodeList clientElems = docElem.getElementsByTagName("client");
		return clientElems.getLength();
	}
	
	private boolean isClientRegistered(String clientID, Element docElem)
	{
		NodeList clientNodes = docElem.getChildNodes();
		int clientsNum = clientNodes.getLength() - 1;
		for(int i = 0; i < clientsNum; i++)
		{
			Element curClientElem = (Element)clientNodes.item(i);
			if(curClientElem.getAttribute("id").equals(clientID))
				return true;
		}
		
		return false;
	}
			
	private boolean shouldIncrementActivity(String serviceName)
	{
		// Return false if this is a rule (e.g., "ContRule_r1") and not a service.
		return serviceName.startsWith("ContService_");
	}
	
	private void updateIdleness(String serviceName, String servicePeer)
	{
		// After receiving a message, we should call IdlenessManager.onRecvMsg
		// for the relation that received the message.
		IdlenessManagerServiceLocator locator = new IdlenessManagerServiceLocator();
		try
		{
			// The service is a local service, since the register service
			// is called on the peer that received the message.
			URL serviceURL = new URL(dQSQUtils.getPeerURL() + "/services/IdlenessManager");
			localhost.axml.services.IdlenessManager.IdlenessManager service = locator.getIdlenessManager(serviceURL);
			service.onRecvMsg(dQSQUtils.service2Relation(serviceName));
		}
		catch(Exception e)
		{
			dQSQUtils.debug("ERROR - RegisterService failed to call IdlenessManager.onRecvMsg for service " + serviceName + ". System error: " + e.getMessage(), "debug_RegisterService.txt");
		}
	}
}
