package il.ac.tau.cs.dqsq.common;

import il.ac.tau.cs.dqsq.qsq2axml.Qsq2AxmlTranslator;
import il.ac.tau.cs.dqsq.util.dQSQUtils;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;
import java.util.HashMap;
import org.jdom.Element;

/**
 * @author pettelno
 *
 * A single relation occurrence in a dDatalog program.
 * That is an instance of a relation in the program. For example, in the
 * program
 * r(x,y) :- a(x,y)
 * s(x,y) :- r(x,z)
 * relation r has 2 occurrences, while relation s has one occurrence.
 * 
 */
public class RelationOccurrence
{
	private static final String REL_NAME = "name"; 
	private static final String REL_ADORNMENT = "adornment"; 
	private static final String REL_PEER_NAME = "peerName"; 
	private static final String REL_PEER_URL = "peerURL"; 
	private static final String REL_VARS= "vars"; 
	
    private String name;
    private List variables = new ArrayList();
    private String adornment = null;
    
    // The URL of the peer that hosts the relation, e.g.
    // http://pc-milo1:8080/axml/servlet/AxisServlet.
    private String serviceURL = null;

    // The URL of the peer that hosts the relation in its short version, e.g.
    // http://pc-milo1:8080/axml.    
    private String serviceShortURL = null;
    
    // The peer that hosts the relation.
    private String peerName = null;
    
    RelationOccurrence()
	{
	}
    
    public RelationOccurrence(String relationName, String peerName,
    						  String peerURL, List variables, String adornment)
    {
    	this.name = relationName;
    	this.variables = variables;
    	this.peerName = peerName;
    	this.adornment = adornment;
    	
    	if(peerName != null)
    		setURL(peerURL);
    }
    
    RelationOccurrence(String relationAsString)
	{
        // Constructs a RelationOccurrence object from a string in the format:
        // parent[bf]@R(x,y)
    	
    	// Omit whitespaces.
    	relationAsString = relationAsString.trim();
    	
    	String[] splittedRelation = relationAsString.split("@");

    	// Set relation name and adornment.
    	boolean adornmentExists = false;
    	int startAdornmentIndex = splittedRelation[0].indexOf('[');
    	if(startAdornmentIndex > -1)
    		adornmentExists = true;
    	if(adornmentExists)
    	{
        	int endAdornmentIndex = splittedRelation[0].indexOf(']');
        	this.adornment = splittedRelation[0].substring(
        						startAdornmentIndex + 1, endAdornmentIndex);
        	this.name = splittedRelation[0].substring(0, startAdornmentIndex);
    	}
    	else // no adornment
    		this.name = splittedRelation[0];
    	
    	// Remove query index.
//    	String[] splittedName = this.name.split("<");
//    	this.name = splittedName[0];
    	
    	// Set peer.
    	int startVariableaIndex = splittedRelation[1].indexOf('(');
    	this.peerName = splittedRelation[1].substring(0, startVariableaIndex);

    	// Set variables.
    	String vars = splittedRelation[1].substring(
    					startVariableaIndex + 1,
						splittedRelation[1].length() - 1);
    	StringTokenizer varsTokenizer = new StringTokenizer(vars, ",");
    	while (varsTokenizer.hasMoreTokens())
    		this.variables.add(varsTokenizer.nextToken().trim());
    	
    	// Set URL. This constructor is used to build  RelationOccurrence
    	// objects always on the current peer.
    	// TODO This is not true!!! Can be also remote peer. Fix it.
    	setURL(dQSQUtils.getPeerURL());
	}

    RelationOccurrence(HashMap relation)
	{
    	this.name = (String)relation.get(REL_NAME);
    	this.adornment = (String)relation.get(REL_ADORNMENT);
    	this.peerName = (String)relation.get(REL_PEER_NAME);
    	this.serviceShortURL= (String)relation.get(REL_PEER_URL);
		
    	setURL((String)relation.get(REL_PEER_URL));
    	setVariables((String[])relation.get(REL_VARS));
	}
    
    RelationOccurrence(RelationOccurrence relOccurrence)
	{
    	this.adornment = relOccurrence.adornment;
    	this.name = relOccurrence.name;
    	this.peerName = relOccurrence.peerName;
    	this.serviceShortURL = relOccurrence.serviceShortURL;
    	this.serviceURL = relOccurrence.serviceURL;
    	cloneVariables(relOccurrence.variables);
	}
	
    void read(Element relationElem)
    {
        name = relationElem.getAttributeValue("name");
        serviceShortURL = relationElem.getChildText("serviceURL");
        serviceURL = serviceShortURL + "/servlet/AxisServlet";
        peerName = relationElem.getChildText("peer");
        List parameterElems = relationElem.getChild("parameters").getChildren();
        Iterator parametersIter = parameterElems.iterator();
        while(parametersIter.hasNext())
        {
            Element parameterElem = (Element)parametersIter.next();
            variables.add(parameterElem.getText());
        }
    }

    public int numberOfAttributes()
    {
        return variables.size();
    }

    public String getAttributeName(int index)
    {
        return (String)variables.get(index);
    }

    public boolean contains(String varName)
    {
        return variables.contains(varName);
    }

    public String getName()
    {
    	return this.name;
    }
    
    public String getAdornedName()
    {
    	// Returns the full name of the relation (with the adornmant, if exists).
    	// E.g., r_bf.
    	if(this.adornment == null)
    		return this.name + "_" + Qsq2AxmlTranslator.getQueryIndex();
//    		return this.name;
    	
    	return this.name + "_" + this.adornment + "_" + Qsq2AxmlTranslator.getQueryIndex();
//    	return this.name + "_" + this.adornment;
    }
	    
    public String getServiceURL()
    {
    	return serviceURL;
    }

    public String getServiceShortURL()
    {
    	return serviceShortURL;
    }
	    
    public String getPeerName()
    {
    	return peerName;
    }
	        
    public void setAdornment(String adornment)
    {
    	this.adornment = adornment;
    }
    
    public String toString()
    {
    	// Returns the relation occurrence as a String in the format:
    	// parent[bf]@R(x,y).
    	
//    	String relationAsString = this.name + "<" + Qsq2AxmlTranslator.getQueryIndex() + ">";
    	String relationAsString = this.name;
    	    		
    	if(this.adornment != null)
    		relationAsString += "[" + this.adornment + "]";

    	if(this.peerName != null)
    		relationAsString += "@" + this.peerName;
    	
    	relationAsString += "(";

    	int numOfVars = numberOfAttributes();
    	for(int index = 0; index < numOfVars; index++)
    	{
    		if(index > 0)
    			relationAsString += ",";
    		
    		relationAsString += this.variables.get(index);
    	}
    	
    	relationAsString += ")";
    	
    	return relationAsString;
    }
    
    List getBoundedVariables(String adornment)
    {
    	// Returns the list of variables in the relation that are bounded
    	// to the adornment given as a parameter.

    	List boundedVars = new ArrayList();
    	
    	int varsNum = this.variables.size();
    	for(int index = 0; index < varsNum; index++)
    	{
    		if(adornment.charAt(index) == 'b')
    			boundedVars.add(this.variables.get(index));
    	}
    	
    	return boundedVars;
    }

    public List getBoundedVariables()
    {
    	// Returns the list of variables in the relation that are bounded
    	// to the adornment in the relation.
    	return getBoundedVariables(this.adornment);
    }

    List getVariables()
    {
    	return this.variables;
    }
    
    public String getAdornment()
    {
    	return this.adornment;
    }

    private void setURL(String peerURL)
    {
//    	this.serviceShortURL = dQSQUtils.getPeerURL(this.peerName);
    	this.serviceShortURL = peerURL;
    	this.serviceURL = this.serviceShortURL + "/servlet/AxisServlet";
    }
    
    public void clearAdornment()
    {
    	if(this.adornment != null)
    		this.adornment = null;
    }
    
    public int getBindingLevel()
    {
    	// The binding level of a relation occurrence is the 
    	// number of its bounded variables, which is the number
    	// of 'b' characters in the adornment.
    	int bindingLevel = 0;
    	int numOfVars = this.adornment.length();
    	for(int index = 0; index < numOfVars; index++)
    	{
    		if(this.adornment.charAt(index) == 'b')
    			bindingLevel++;
    	}
    	
		return bindingLevel;
    }
    
    public void concatName(String str)
    {
    	this.name += str;
    }
    
    private void setVariables(String[] varsArray)
    {
    	for(int varIndex = 0; varIndex < varsArray.length; varIndex++)
    		this.variables.add(varsArray[varIndex]);
    }
    
    HashMap toHashMap()
    {
    	HashMap hash = new HashMap();
    	hash.put(REL_NAME, this.name);
    	hash.put(REL_ADORNMENT, this.adornment);
    	hash.put(REL_PEER_NAME, this.peerName);
    	hash.put(REL_PEER_URL, this.serviceShortURL);
    	hash.put(REL_VARS, getVarsInArray());
    	
    	return hash;
    }
    
    private String[] getVarsInArray()
    {
    	int varsNum = numberOfAttributes();
    	String[] varsArray = new String[varsNum];

    	for(int varIndex = 0; varIndex < varsNum; varIndex++)
    		varsArray[varIndex] = (String)this.variables.get(varIndex);
    		
    	return varsArray;
    }
    
    private void cloneVariables(List variables)
    {
    	int size = variables.size();
    	for(int varIndex = 0; varIndex < size; varIndex++)
    		this.variables.add(varIndex, new String((String)variables.get(varIndex)));
    }
}
