package il.ac.tau.cs.dqsq.common;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import org.jdom.Element;
import java.util.StringTokenizer;
import java.util.HashMap;

/**
 * @author pettelno
 *
 * A dDatalog rule.
 * 
 */
public class Rule
{
	private static final String RULE_HEAD = "head"; 
	private static final String RULE_BODY = "body"; 
	
	protected String name = "";
	protected RelationOccurrence head = null;
    
	// A List of RelationOccurrence objects, each representing a relation
	// in the body of the rule.
	protected List body = new ArrayList(); 

    public Rule(Rule rule)
    {
    	this.name = rule.name;
    	this.head = new RelationOccurrence(rule.head);
    	cloneBody(rule.body);
    }
    
    public Rule()
    {
    	this.head = new RelationOccurrence();
    }
    
    public Rule(RelationOccurrence head, List body)
    {
    	this.head = head;
    	this.body = body;
    }

    public Rule(String ruleAsString)
    {    	
        // Rules are sent from peer to peer as Strings. When a peer receives
        // a rule as a String, it constructs a Rule object with this constructor.
        // Rule format as a String:
    	// sup_r1_bf_1@R(x,y) :- sup_r1_bf_0@P(x), parent[bf]@R(x,y)

    	String[] splittedRule = ruleAsString.split(":-");
    	this.head = new RelationOccurrence(splittedRule[0]);

    	buildRuleBody(splittedRule[1]);
    }
    
    public Rule(HashMap ruleAsHashMap)
    {
    	this.head = new RelationOccurrence(
    			(HashMap)ruleAsHashMap.get(RULE_HEAD));
    	buildRuleBody((HashMap[])ruleAsHashMap.get(RULE_BODY));
    }
    
    public Rule(Element ruleElem)
    {
    	this.head = new RelationOccurrence();
    	read(ruleElem);
    }
    
    public void read(Element ruleElem)
    {
        name = ruleElem.getAttributeValue("name");
        readHead(ruleElem.getChild("head"));
        readBody(ruleElem.getChild("body"));
    }

    private void readHead(Element headElem)
    {
        Element relationElem = headElem.getChild("relation");
        head.read(relationElem);
    }

    private void readBody(Element bodyElem)
    {
        List relationElems = bodyElem.getChildren();
        Iterator relationsIter = relationElems.iterator();
        while(relationsIter.hasNext())
        {
            Element relationElem = (Element)relationsIter.next();
            RelationOccurrence curRelation = new RelationOccurrence();
            curRelation.read(relationElem);
            body.add(curRelation);
        }
    }
    
    public String getHeadRelationName()
    {
    	return this.head.getName();
    }

    public String getHeadRelationAdornedName()
    {
    	return this.head.getAdornedName();
    }
    
    public List getBody()
    {
    	return this.body;
    }
    
    public String getName()
    {
    	return this.name;
    }
    
    public RelationOccurrence getRelationFromBody(int index)
    {
    	// Index starts from 0, i.e. a call to this method with index=0
    	// returns the first relation in the rule body.
    	return (RelationOccurrence)this.body.get(index);
    }
    
    public int getHeadsNumOfAttributes()
    {
    	return head.numberOfAttributes();
    }
    
    public String getHeadsAttributeName(int index)
    {
    	return head.getAttributeName(index);
    }
    
    public String getBodysRelationName(int index)
    {
    	// Index starts from 0, i.e. a call to this method with index=0
    	// returns the name of the first relation in the rule body.
    	return ((RelationOccurrence)body.get(index)).getName();
    }

    public String getBodysRelationAdornedName(int index)
    {
    	// Index starts from 0, i.e. a call to this method with index=0
    	// returns the name of the first relation in the rule body.
    	return ((RelationOccurrence)body.get(index)).getAdornedName();
    }
    
    public boolean isSingleRelationInBody()
    {
    	return (this.body.size() == 1);
    }
    
    public String toString()
    {
        // Rules are sent from peer to peer as Strings. When a peer sends
        // a rule it uses this method to transform it to a String. 

    	String ruleAsString = this.head.toString();
       	ruleAsString += " :- ";
       	
    	int numOfRelationsInBody = this.body.size();
    	for(int index = 0; index < numOfRelationsInBody; index++)
    	{
    		if(index > 0)
    			ruleAsString += ", ";
    		
    		ruleAsString += getRelationFromBody(index).toString();
    	}
       	
       	return ruleAsString;
    }
    
    public int getNumOfBodyRelations()
    {
    	return this.body.size();
	
    }
    
    public String getHostingPeer()
    {
    	return this.head.getPeerName();
    }

    public List getHeadVariables()
    {
    	return this.head.getVariables();
    }

    public List getHeadBoundedVariables(String adornment)
    {
    	// Returns the list of variables in the rule head that are bounded
    	// to the adornment given as a parameter.
    	return this.head.getBoundedVariables(adornment);
    }
    
    public List getBodyVariablesAtIndex(int index)
    {
    	// Index starts from 0, i.e. a call to this method with index=0
    	// returns the variables of the first relation in the rule body.
    	return ((RelationOccurrence)this.body.get(index)).getVariables();
    }
    
    public RelationOccurrence getHeadRelation()
    {
    	return this.head;
    }

    public void setName(String name)
    {
    	this.name = name;
    }
    
    public String getHeadShortURL()
    {
    	return this.head.getServiceShortURL();
    }
  
    private void buildRuleBody(String bodyAsString)
    {
        // Rule body format: sup_r1_bf_0@P(x), parent[bf]@R(x,y)

    	StringTokenizer bodyTokenizer = new StringTokenizer(bodyAsString, ")");
    	String currentRelation = null;
    	boolean isFirstRelation = true;
    	while (bodyTokenizer.hasMoreTokens())
    	{
    		currentRelation = bodyTokenizer.nextToken();
    		if(!isFirstRelation)
    			// Remove the ',' character from the beginning of the string.
    			currentRelation = currentRelation.substring(1);
    		else
    			isFirstRelation = false;
    		
    		currentRelation = currentRelation.concat(")");
    		
        	this.body.add(new RelationOccurrence(currentRelation));
    	}	    	
    }
    
    private void buildRuleBody(HashMap[] bodyAsHashMapArray)
    {
    	for(int relIndex = 0; relIndex < bodyAsHashMapArray.length; relIndex++)
    		this.body.add(new RelationOccurrence(bodyAsHashMapArray[relIndex]));
    }
    
	/**
	 * @param relationsNames A Set of relation names.
	 * @return The relations in relationsNames which also appear in
	 *  	   in the rule body.
	 */
	public Set matchBodyRelations(Set relationsNames)
	{
		Set intensionalRelations = new HashSet();

		int numOfRelationsInBody = getNumOfBodyRelations();
		for(int index = 0; index < numOfRelationsInBody; index++)
		{
			RelationOccurrence relation = getRelationFromBody(index);
			if(relationsNames.contains(relation.getName()))
				intensionalRelations.add(relation);
		}
		
		return intensionalRelations;
	}
	    
    public void setBodyRelationAdornment(int index, String adornment)
    {
    	// Index starts from 0, i.e. a call to this method with index=0
    	// returns the name of the first relation in the rule body.
    	((RelationOccurrence)this.body.get(index)).setAdornment(adornment);
    }
        
    protected void switchRelationsInBody(int relation1Index, int relation2Index)
    {
    	RelationOccurrence tmpRelation = getRelationFromBody(relation1Index);
    	this.body.set(relation1Index, this.body.get(relation2Index));
    	this.body.set(relation2Index, tmpRelation);
    }
    
/*    
    public void concatRelationNames(String str)
    {
    	this.head.concatName(str);

    	int numOfRelationsInBody = getNumOfBodyRelations();
		for(int index = 0; index < numOfRelationsInBody; index++)
		{
			RelationOccurrence relation = getRelationFromBody(index);
			relation.concatName(str);
		}
    }
*/
    
    public HashMap toHashMap()
    {
    	HashMap ruleAsHashMap = new HashMap();
    	ruleAsHashMap.put(RULE_HEAD, this.head.toHashMap());
    	ruleAsHashMap.put(RULE_BODY, getBodyAsHashMapArray());
    	
    	return ruleAsHashMap;
    }
    
    private HashMap[] getBodyAsHashMapArray()
    {
    	int numOfBodyRelations = getNumOfBodyRelations();
    	HashMap[] body = new HashMap[numOfBodyRelations];
    	
    	for(int relIndex = 0; relIndex < numOfBodyRelations; relIndex++)
    		body[relIndex] = getRelationFromBody(relIndex).toHashMap();
    	
    	return body;
    }
    
    public String getBodysRelationPeerName(int index)
    {
    	// Index starts from 0, i.e. a call to this method with index=0
    	// returns the variables of the first relation in the rule body.
    	return ((RelationOccurrence)this.body.get(index)).getPeerName();
    }
    
    private void cloneBody(List body)
    {
    	int size = body.size();
    	for(int relIndex = 0; relIndex < size; relIndex++)
    		this.body.add(
    				relIndex,
					new RelationOccurrence((RelationOccurrence)body.get(relIndex)));    	
    }
}
