package il.ac.tau.cs.dqsq.datalog2qsq;

import il.ac.tau.cs.dqsq.common.Relation;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import org.jdom.Element;

/**
 * @author pettelno
 *
 * A container for a Datalog program.
 * 
 */
class DatalogProgram
{
    private Map extensionalRelations = new HashMap();
    private Set intensionalRelations = new HashSet();
    private Collection rules = new ArrayList();
    
	DatalogProgram(Element root)
	{
		// Receives a JDom Element representing the root of the XML file
		// that stores the Datalog program (datalog.xml).
		readInput(root);
	}	
	
	private void readInput(Element root)
	{
	    // Read relations.
	    Element relationsElem = root.getChild("extensionalRelations");
	    if(relationsElem != null)
	        readRelations(relationsElem);
	
	    // Read rules.
	    Element rulesElem = root.getChild("rules");
	    if(rulesElem != null)
	        readRules(rulesElem);
	}
	
    private void readRelations(Element relationsElem)
    {
        List relationElems = relationsElem.getChildren();
        Iterator relationsIter = relationElems.iterator();
        while(relationsIter.hasNext())
        {
            Element relationElem = (Element)relationsIter.next();
            Relation curRelation = new Relation();
            curRelation.read(relationElem);
            extensionalRelations.put(curRelation.getShortName(), curRelation);
        }
    }

    private void readRules(Element rulesElem)
    {
        List ruleElems = rulesElem.getChildren();
        Iterator rulesIter = ruleElems.iterator();
        while(rulesIter.hasNext())
        {
            Element ruleElem = (Element)rulesIter.next();
            RewritableRuleManager curRule = new RewritableRuleManager(ruleElem);
//            RewritableRule curRule = new RewritableRule();
//            curRule.read(ruleElem);
            rules.add(curRule);
            this.intensionalRelations.add(curRule.getHeadRelationName());
        }
    }
        
    Collection getRulesOfIntensionalRelation(String relationName)
    {
    	// Returns all the rules in which the relation in the head is
    	// the parameter relationName i.e., the rules that define the
    	// given intensional relation.
    	
    	Collection retRules = new ArrayList();
    	
    	Iterator iter = this.rules.iterator();
    	while(iter.hasNext())
    	{
    		RewritableRuleManager rule = (RewritableRuleManager)iter.next();
    		if(rule.getHeadRelationName().equals(relationName))
    			retRules.add(rule);
    	}
    	
    	return retRules;
    }
    
    Map getExtensionalRelations()
    {
    	return this.extensionalRelations;
    }
    
    Set getIntensionalRelations()
    {
    	return this.intensionalRelations;
    }
    
    void clearAdornmentsHistory()
    {
    	Iterator rulesIter = this.rules.iterator();
    	while(rulesIter.hasNext())
    	{
    		RewritableRuleManager rule = (RewritableRuleManager)rulesIter.next();
    		rule.clearAdornmentsHistory();
    	}
    }
    
    HashMap serializeProgram()
    {
    	// Returns the Datalog program as a HashMap with 2 entries:
    	// "rules" and "relations".
    	// "rules" is an ArrayList of the rules as Strings.
    	// "relations" is a HashMap of ArrayLists, each contains an ArrayList
    	// of the tuples that exist in the extensional relation.
    	HashMap program = new HashMap();
    	program.put("rules", serializeRules());
    	program.put("relations", serializeExtensionalRelations());
    	return program;
    }

/*    
    private HashMap serializeRules()
    {
    	HashMap rules = new HashMap();
    	Iterator rulesIter = this.rules.iterator();
    	while(rulesIter.hasNext())
    	{
    		RewritableRule rule = (RewritableRule)rulesIter.next();
    		rules.put(rule.getName(), rule.toString());
    	}
    	
    	return rules;
    }   
*/
    

    private ArrayList serializeRules()
    {
    	ArrayList rules = new ArrayList();
    	Iterator rulesIter = this.rules.iterator();
    	while(rulesIter.hasNext())
    	{
    		RewritableRuleManager inRule = (RewritableRuleManager)rulesIter.next();
    		HashMap outRule = new HashMap();
    		outRule.put("name", inRule.getName());
    		outRule.put("rule", inRule.toString());
    		rules.add(outRule);
    	}
    	
    	return rules;
    }
    
    private HashMap serializeExtensionalRelations()
    {
    	HashMap relations = new HashMap();

    	Iterator relationsIter = this.extensionalRelations.keySet().iterator();
    	while(relationsIter.hasNext())
    	{
    		String relationName = (String)relationsIter.next();
    		Relation relation = (Relation)this.extensionalRelations.get(relationName);
    		relations.put(relationName, relation.serialize());
    	}
    	
    	return relations;
    }
}
