package il.ac.tau.cs.dqsq.datalog2qsq;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;
import il.ac.tau.cs.dqsq.qsq2axml.Qsq2AxmlTranslator;
import java.io.File;
import java.net.URL;
import java.util.Iterator;
import il.ac.tau.cs.dqsq.util.dQSQUtils;
import il.ac.tau.cs.dqsq.common.*;
import java.util.Collection;
import il.ac.tau.cs.dqsq.util.Logger;
//import localhost.axml.services.Datalog2QSQ.Datalog2QSQSoapBindingStub;
import localhost.axml.services.Datalog2QSQ.PeerServiceLocator;
import java.util.Set;
import java.util.List;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.HashMap;

/**
 * @author pettelno
 *
 * A peer in the dQSQ system. It provides the entry points to the system
 * by defining web services that includes the following:
 * 1. runQuery - Receives a query and starts the dQSQ computation process. 
 * 2. recvRuleFromRemotePeer - While the dQSQ rewriting is performed, a 
 *    peer uses this service to send a rule to a remote peer.
 * 3. terminateQuery - Clears the system from the current query.
 * 4. getDatalogProgram - Returns the original Datalog program serialized
 *    in a HashMap. Used by the JSP UI.
 * 5. sendLogMessage - Sends a message to the log. Used by the JSP UI.
 * 
 * Each peer reads its own part of the dDatalog program from the input file
 * datalog.xml, which is located in the AXML repository. One of the peers
 * then receives a query. The peer that receives the query is the one that
 * starts the dQSQ computation process, as described in the article
 * "Diagnosis of Asynchronous Discrete Event Systems: Datalog to the Rescue!"
 * by Abiteboul et al. One slight difference of the algorithm implemented here
 * from the algorithm in the article, is that in the latter, when a peer 
 * encounters a remote relation while rewriting a rule, it delegates the
 * processing of the remainder of the rule to the remote peer in charge of
 * that relation. In the algorithm implemented here, every rule is rewritten
 * completely by a single peer. After the rewriting of the rule is finished,
 * the rewriting peer sends every supplementary rule to the peer that should
 * be in charge of that rule. For example, if the original rule is
 * r@R(x,y) :- s@S(x,z), t@T(z,y), then peer R rewrites it. It knows in
 * advance that the sup0 rule should remain local, the sup1 rule should be
 * sent to peer S and the sup2 rule should be sent to peer T.
 * 
 * The dQSQ system can run several queries, which are issued sequentially.
 * Simultaneous queries are not supported.
 * 
 * Each Peer object holds a Qsq2AxmlTranslator object, which acts as a
 * gate to the qsq2axml package and used for the physical translation from
 * QSQ format to AXML format.
 *   
 */
public class Peer
{
	private String peerName = null;
	private String peerURL = null;
	private DatalogProgram inputDatalogProg = null;
	private Qsq2AxmlTranslator qsq2Axml = null;
    private String repositoryLocation = null;
    private Query query = null;
    private boolean isActive = false; 
    private boolean isTerminatingQuery = false; 
    
    // Stores the URLs of the remote peers which take part of the current
    // query computation.
    private Set remotePeersURLs = new HashSet();

    // The original Datalog program serialized into a HashMap.
    // Needed for displaying the Datalog program in the UI.
    private HashMap serializedDatalogProgram = null;
	
    private List outputQsqRules = new ArrayList();
	private long queryStartTime;
	private long queryEndTime;
    
    /*
     * Constructs a new Peer object. This constructor is autoamatically
     * called by Axis when the first Web service in this class is invoked.
     * If the query is issued on this peer, then the first call is to
     * runQuery, otherwise the first call is to recvRuleFromRemotePeer.
     * The constructor loads the dDatalog program and stores it in memory.
     * 
     */
    public Peer()
    {
    	startDQSQ();
    }
    
	private void startDQSQ()
	{
		this.repositoryLocation = dQSQUtils.getRepositoryLocation();
		String inputFileName = this.repositoryLocation + File.separator + "datalog.xml";
		readInput(inputFileName);
		
		Logger.init(this.peerName);
		Logger.sendMessage("Started dQSQ process.", Logger.LOG_LEVEL_DEBUG);        	
		
		initQsq2AxmlTranslator();
		
		// The extensional relations can be translated in advance.
		// This translation is valid for all the subsequent queries.
//		translateExtensionalRelations();

		// Old code that received the query as a part of the input.
//		if(this.query != null)
//			processQuery();
	}

	private void readInput(String inputFileName)
	{
        // Build the document with SAX and Xerces, no validation.
        SAXBuilder builder = new SAXBuilder();
        
        try
		{
        	Document doc = builder.build(new File(inputFileName));
        
        	Element root = doc.getRootElement();

        	// Read general settings.
        	this.peerName = root.getAttributeValue("peer");
        	this.peerURL = dQSQUtils.getPeerURL();
        	
        	this.inputDatalogProg = new DatalogProgram(root);
        	
        	// this.serializedDatalogProgram is initialized now and not
        	// lazily in getDatalogProgram, because we want the Datalog
        	// program with no adornments. If initialized in getDatalogProgram,
        	// it may contain adornments.
        	this.serializedDatalogProgram = this.inputDatalogProg.serializeProgram();
        	this.serializedDatalogProgram.put("peerName", this.peerName);
        	
    		// Old code that received the query as a part of the input.
//        	Element queryElem = root.getChild("query");
//        	if(queryElem != null)
//        	{
//        		Logger.debug("Received a query.");        	
//        		this.query = new Query(queryElem);
//        	}
		}
        catch(Exception e)
		{
        	Logger.sendMessage("Failed to read input.", Logger.LOG_LEVEL_ERROR);        	
		}
	}
    
	private void initQsq2AxmlTranslator()
	{
		try
		{
			this.qsq2Axml = new Qsq2AxmlTranslator(repositoryLocation, this.peerName);
		}
		catch(Exception e)
		{
			Logger.sendMessage("Failed to initialize QSQ2AXML translator.",
							   Logger.LOG_LEVEL_ERROR);
		}

	}

	private void processQuery()
	{
		setQueryAdornedRelation();
		setQueryInputRelation();
		setQueryRule();
		rewriteRules4IntensionalRelelation(this.query.getBodyRelationName(), this.query.getAdornment());
	}
	
	private void setQueryAdornedRelation()
	{
		// In the example from Petri Nets article: r_bf.  
		
		Relation adornedRel = this.query.setAdornedRelation();
		
		// This relation is the leader for termination detection,
		// unless termination detection is disabled.
		boolean isLeader = false;
		if(dQSQUtils.isTerminationDetectionOn())
			isLeader = true;
		this.qsq2Axml.translateRelation(adornedRel, isLeader);
	}

	private void setQueryInputRelation()
	{
		// In the example from Petri Nets article: in_r_bf.  

		Relation inputRel = this.query.setInputRelation();
		this.qsq2Axml.translateRelation(inputRel, false);
	}
	
    private void rewriteRules4IntensionalRelelation(String relationName, String adornment)
    {
    	// Rewrites all the rules that define the given relation.
    	// This method is first used with the relation in the body of
    	// the query. In the example from Petri Nets article: Rules 1 and
    	// 2 that define relation r.
    	// The method is also used later, for recursive rewriting with
    	// relations that are ecountered during the query computation.
    	
    	Logger.sendMessage("Rewriting rules for intensional relation " + relationName +
    				 	   " with adornment " + adornment + ".",
						   Logger.LOG_LEVEL_FLOW);
		
		Collection rules = this.inputDatalogProg.getRulesOfIntensionalRelation(relationName);
    	Iterator iter = rules.iterator();
    	while(iter.hasNext())
    	{
    		RewritableRuleManager ruleManager = (RewritableRuleManager)iter.next();
    		rewriteRule(ruleManager, adornment);
    	}
    }
    
    /*synchronized*/ private void rewriteRule(RewritableRuleManager ruleManager, String adornment)
    {
        // This method is synchronized. Simultaneous rule rewritings are not
        // allowed, since we want to prevent simultaneous rewriting of the
        // same rule (with different adornments). The reason for this is that
        // each rewriting uses the adornment fields of the RelationOccurence
        // objects in the rule.    	
    	
    	Logger.sendMessage("----------------------------------------------",
    					   Logger.LOG_LEVEL_DEBUG);
    	Logger.sendMessage("Rewriting rule " + ruleManager.getName() + ".",
    					   Logger.LOG_LEVEL_DEBUG);
		
		QsqRewriting qsqRewriting = new QsqRewriting(
				ruleManager, adornment, 
				this.inputDatalogProg.getIntensionalRelations(),
				this.inputDatalogProg.getExtensionalRelations());
		if(!qsqRewriting.isEmpty())
		{
			translateQsqRules(qsqRewriting);		
		
			// Rewrite all local rules that define the relations that
			// appear in the rule.
			// In the example from Petri Nets article: If all 4 rules are
			// located on peer R, then after rewriting rule 2, rules 3 and 4
			// should be rewritten.
			rewriteRecursively(qsqRewriting.getOriginalRuleAfterSort());
		}
    }
    
    private void translateQsqRules(QsqRewriting qsqRewriting)
    {
    	// The order of translation must be as follows:
    	// 1. Local relations.
    	// 2. Remote rules (and relations).
    	// 3. Local rules.
    	// This order must be kept in order to avoid synchronization
    	// problems between different peers. The idea is that a rule
    	// should be translated only after all the relations that
    	// appear in the rule are translated. A remote rule may contain
    	// a call to a local relation, thus the local relations are
    	// translated first. Similarly, a local rule may contain
    	// a call to a remote relation, thus the remote relations are
    	// translated first.
    	// Note: To translate a remote rule, the local peer sends
    	// the rule to the remote peer, by calling a web service.
    	// Calling a web service is synchronous, which assures us
    	// that the local rules are translated only after the translation
    	// of the remote rules is finished. Cool!
    	translateLocalRelations(qsqRewriting);    	
    	translateRemoteRules(qsqRewriting);
    	translateLocalRules(qsqRewriting);    	
    }    

    private void translateLocalRelations(QsqRewriting qsqRewriting)
    {
    	// The sup0 rule and the main rule are always local.
    	// The "main rule" is the rule that defines the relation in the head
    	// of the original rule from the last sup rule. For example, if the
    	// original rule is r(x,y) :- s(x,z), t(z,y) then the main rule is
    	// r(x,y) :- sup2(x,y).
    	translateLocalRelationsInRule(qsqRewriting.getSup0());
    	translateLocalRelationsInRule(qsqRewriting.getMainRule());    	
    	
    	int numOfSupRules = qsqRewriting.numOfSupRules();
    	for(int index = 1; index < numOfSupRules; index++)
    	{
    		Rule rule = qsqRewriting.getSupi(index);
    		if(rule.getHostingPeer().equals(this.peerName))
    			translateLocalRelationsInRule(rule);
    	}
    }

    private void translateRemoteRules(QsqRewriting qsqRewriting)
    {
    	int numOfSupRules = qsqRewriting.numOfSupRules();
    	for(int index = 1; index < numOfSupRules; index++)
    	{
    		Rule rule = qsqRewriting.getSupi(index);
    		if(!rule.getHostingPeer().equals(this.peerName))
    			translateRemoteRule(rule);
    	}
    }

    private void translateLocalRules(QsqRewriting qsqRewriting)
    {
    	// The sup0 rule and the main rule are always local.
    	translateLocalRule(qsqRewriting.getSup0());
    	translateLocalRule(qsqRewriting.getMainRule());    	
    	
    	int numOfSupRules = qsqRewriting.numOfSupRules();
    	for(int index = 1; index < numOfSupRules; index++)
    	{
    		Rule rule = qsqRewriting.getSupi(index);
    		if(rule.getHostingPeer().equals(this.peerName))
    		{
    			translateLocalRule(rule);
    			translateInputSupRule(rule);    			
    		}
    	}
    }

    private void translateLocalRelationsInRule(Rule rule)
    {
    	// 1. Translate relation in the rule head.
		this.qsq2Axml.translateRelation(new Relation(rule.getHeadRelation()), false);
    	
    	// 2. Translate all local relations in the rule body.
    	int numOfBodyRelations = rule.getNumOfBodyRelations();
    	for(int index = 0; index < numOfBodyRelations; index++)
    	{
    		RelationOccurrence relOccurrence = rule.getRelationFromBody(index);
    		if(relOccurrence.getPeerName().equals(this.peerName))
    			this.qsq2Axml.translateRelation(new Relation(relOccurrence), false);
    	}    	
    }

    private void translateLocalRule(Rule rule)
    {
    	this.outputQsqRules.add(rule.toString());
    	this.qsq2Axml.translateRule(rule);
    }

    private void translateRemoteRule(Rule rule)
    {
    	// Sends the rule as a HashMap object to the remote peer via the
    	// web service recvRuleFromRemotePeer.
    	
    	Logger.sendMessage("Sending rule " + rule.toString() + " to remote peer " +
    					   rule.getHostingPeer() + ".",
    					   Logger.LOG_LEVEL_DEBUG);

    	PeerServiceLocator locator = new PeerServiceLocator();
		try
		{
			URL serviceURL = new URL(rule.getHeadShortURL() + "/services/Datalog2QSQ");
			this.remotePeersURLs.add(serviceURL);
			localhost.axml.services.Datalog2QSQ.Peer remotePeer = locator.getDatalog2QSQ(serviceURL);
			HashMap ruleAsHash = rule.toHashMap();
			ruleAsHash.put("queryID", Qsq2AxmlTranslator.getQueryIndex());
			remotePeer.recvRuleFromRemotePeer(ruleAsHash);
		}
		catch(Exception e)
		{
			Logger.sendMessage("Failed to send rule " + rule.toString() + " to remote peer " + rule.getHostingPeer() +
					     	   " (system error: " + e.getMessage() + ".",
							   Logger.LOG_LEVEL_ERROR);
		}    	    			
    } 
        
    /**
     * This method is exposed as a web service. While the dQSQ rewriting is
     * performed, a peer uses this service to send a rule to a remote peer.
     * 
     * @param ruleAsHashMap A dDatalog rule stored in a HashMap object.
     * 	      See Rule.toHashMap for details about the way a rule is
     * 		  stored in the HashMap. 
     * 
     */
    public void recvRuleFromRemotePeer(HashMap ruleAsHashMap)
    {
    	if(!this.isActive)
    	{
    		// This is the first message that this peer received in
    		// the current query computation.
    		this.isActive = true;
        	Qsq2AxmlTranslator.setQueryIndex((String)ruleAsHashMap.get("queryID"));
    		translateExtensionalRelations();
    	}

    	RewritableRule rule = new RewritableRule(ruleAsHashMap);

    	Logger.sendMessage("Received rule " + rule.toString() + " from remote peer.",
    					   Logger.LOG_LEVEL_DEBUG);
    	
    	// Remove adornmants of local extensional relations. This is
    	// necessary because the rule was sent fully-adorned, since the
    	// sending peer cannot know if a relation in the rule is 
    	// extensional or intensional.
    	rule.removeExtraAdornments(this.inputDatalogProg.getExtensionalRelations());
    	Logger.sendMessage("Rule after removing extra afornments: " + rule.toString() + ".",
    					   Logger.LOG_LEVEL_FLOW);
    	   	
    	translateLocalRelationsInRule(rule);
    	translateLocalRule(rule);

    	translateInputSupRule(rule);
    	
		// Rewrite all local rules that define the relations that
    	// appear in the rule.
		rewriteRecursively(rule);
    }
    
    private void rewriteRecursively(Rule rule)
	{
   	 // Rewrites recursively all the rules that define the local intensional
	 // relations that appear in the rule that is given as a parameter.

	 Set intensionalRelationsInRule =
			rule.matchBodyRelations(
					this.inputDatalogProg.getIntensionalRelations());
		Iterator relationsIter = intensionalRelationsInRule.iterator();
		while(relationsIter.hasNext())
		{			
			RelationOccurrence relation = (RelationOccurrence)relationsIter.next();
	    	Logger.sendMessage("Started recursive rewriting for relation " + relation.getAdornedName() + ".",
	    					   Logger.LOG_LEVEL_FLOW);
			rewriteRules4IntensionalRelelation(relation.getName(), relation.getAdornment());
		}
	}
    
	private void translateInputSupRule(Rule origSupRule)
	{
	    // For each sup rule there is potentially a corresponding input rule.
	    // For example, the input rule for up_r2_bf_1@P(z,x) :- sup_r2_bf_0@P(x), ancestor[bf]@P(x,z)
	    // is in_ancestor[bf]@P(x) :- sup_r2_bf_0@P(x).
	    // Such an input rule is necessary if 1) The original rule has
	    // 2 relations in its body and 2) The second relation in its body
	    // is intensional.
		
		if(shouldTranslateInputRule(origSupRule))
		{
			Rule inputRule = setInputRule(origSupRule);
	    	Logger.sendMessage("Input rule for rule " + origSupRule.getName() +
	    				 	   " is " + inputRule.toString() + ".",
							   Logger.LOG_LEVEL_FLOW);
	    	translateLocalRelationsInRule(inputRule);
	    	translateLocalRule(inputRule);
		}
	}
	
	private boolean shouldTranslateInputRule(Rule origSupRule)
	{
		// Returns true if 1) origSupRule has (at least) 2 relations in its
		// body and 2) The second relation in its body is intensional.
		
		if(origSupRule.getNumOfBodyRelations() < 2)
			return false;
		
		RelationOccurrence secondRelationInBody = origSupRule.getRelationFromBody(1);
		if(isIntensionalRelation(secondRelationInBody.getName()))
			return true;
		
		return false;		
	}
	
	private boolean isIntensionalRelation(String relationName)
	{
		if(this.inputDatalogProg.getIntensionalRelations().contains(relationName))
			return true;
		
		return false;
	}
	
	private Rule setInputRule(Rule origSupRule)
	{
		// Example: The input rule for up_r2_bf_1@P(z,x) :- sup_r2_bf_0@P(x), ancestor[bf]@P(x,z)
	    // is in_ancestor[bf]@P(x) :- sup_r2_bf_0@P(x).		
		
		RelationOccurrence intensionalRelation = origSupRule.getRelationFromBody(1);
		String headName = "in_" + intensionalRelation.getName();
		String headAdornment = intensionalRelation.getAdornment();
		List headVars = intensionalRelation.getBoundedVariables();
		
		RelationOccurrence head = new RelationOccurrence(
				headName, this.peerName, this.peerURL, headVars, headAdornment);
		
		// The only relation in the body of the input rule is the
		// sup relation in the body of the original rule, which
		// is the first relation in the body of the original rule.
		RelationOccurrence bodyRelation = origSupRule.getRelationFromBody(0);
		List body = new ArrayList();
		body.add(bodyRelation);
		
		return new Rule(head, body);
	}
	
    private void translateExtensionalRelations()
	{
		Logger.sendMessage("Translating extensional relations.",
						   Logger.LOG_LEVEL_FLOW);        	
    	
    	Collection extRelationsVals = this.inputDatalogProg.getExtensionalRelations().values();
    	Iterator iter = extRelationsVals.iterator();
    	while(iter.hasNext())
    	{
    		Relation rel = (Relation)iter.next();

    		// Translate each extensional relation to AXML.
            this.qsq2Axml.translateRelation(rel, false);
    	}
	}

    private void setQueryRule()
    {
    	// The query is treated as a regular rule.
    	Rule queryRule = this.query.getQueryRule();
    	translateLocalRelationsInRule(queryRule);
    	translateLocalRule(queryRule);    	
    }
    
    /**
     * This method is exposed as a web service. It receives a query and
     * starts the distributed query computation process.
     * 
     * @param queryAsString A dDatalog query as a String in the format
     *        q@P(y) :- ancestor@P("Joyce", y) 
     */
    public void runQuery(String queryAsString) throws Exception
    {
		this.queryStartTime = System.currentTimeMillis();
    	
    	Logger.sendMessage("==============================================",
				   		   Logger.LOG_LEVEL_DEBUG);
		Logger.sendMessage("Received query " + queryAsString + ".",
						   Logger.LOG_LEVEL_DEBUG);        	

		this.query = new Query(queryAsString, this.inputDatalogProg.getExtensionalRelations(), this.peerName);    	
		this.isActive = true;
		translateExtensionalRelations();		
		processQuery();
    }

    /**
     * This method is exposed as a web service.
     * It clears the system from the current query. This includes removing
     * and unregistering all the AXML documents and services that were
     * generated as a result of the computation of the current query.
     * The method recursively clears the state of all the remote peers
     * which are involved in the computation of the current query. 
     *
     */
    public void terminateQuery()
    {
    	// Check if the peer is currently terminating the query, to
    	// avoid an infinite loop of calls between peers. For example,
    	// peer P is in terminateQuery, then it calls terminateQuery on
    	// peer R, and later R calls terminateQuery on peer P...
    	if(this.isTerminatingQuery)
    		return;
    	
    	this.isTerminatingQuery = true;
    		
		Logger.sendMessage("Terminating query.", Logger.LOG_LEVEL_DEBUG); 

		this.inputDatalogProg.clearAdornmentsHistory();
		this.qsq2Axml.clearPeer();

		String inputFileName = this.repositoryLocation + File.separator + "datalog.xml";
		readInput(inputFileName);
		
//		translateExtensionalRelations();
		
		Iterator remotePeersIter = this.remotePeersURLs.iterator();
		while(remotePeersIter.hasNext())
		{
			URL remotePeerURL = (URL)remotePeersIter.next();
			clearRemotePeer(remotePeerURL);
		}

		this.remotePeersURLs.clear();
		this.isActive = false;
    	this.isTerminatingQuery = false;
    	this.outputQsqRules.clear();
    }
    
    private void clearRemotePeer(URL remotePeerURL)
    {
    	PeerServiceLocator locator = new PeerServiceLocator();
		try
		{
			localhost.axml.services.Datalog2QSQ.Peer remotePeer = locator.getDatalog2QSQ(remotePeerURL);
			remotePeer.terminateQuery();
		}
		catch(Exception e)
		{
			Logger.sendMessage("Failed to clear remote peer at URL " + remotePeerURL + 
					     	   " (system error: " + e.getMessage() + ".",
							   Logger.LOG_LEVEL_ERROR);
		}    	    			    	
    }

    /**
     * This method is exposed as a web service.
     * It returns the original Datalog program serialized in a HashMap.
     * 
     * @return The original Datalog program serialized in a HashMap.
     */
    public HashMap getDatalogProgram()
    {
    	return this.serializedDatalogProgram;
    }

    public List getQsqRewritingOutput()
    {
    	return this.outputQsqRules;
    }

    public String getPeerName()
    {
    	return this.peerName;
    }
    
    /**
     * This method is exposed as a web service.
     * It sends a message to the peer log.
     * 
     * @param message The message to send.
     * @param severity The severity of the message. (See class Logger for
     *        more details).
     * 
     */
    public void sendLogMessage(String message, int severity)
    {
    	Logger.sendMessage(message, severity);
    }

    /**
     * This method is exposed as a web service.
     *
     */
    public void queryComputationFinished()
    {
    	Logger.sendMessage("Query computation is finished.", Logger.LOG_LEVEL_DEBUG);
		this.queryEndTime = System.currentTimeMillis();
    }
    
    /**
     * This method is exposed as a web service.
     *
     */
    public long getQueryComputationTime()
    {
    	return this.queryEndTime - this.queryStartTime; 
    }    
}
