package il.ac.tau.cs.dqsq.datalog2qsq;

import java.util.List;
import il.ac.tau.cs.dqsq.common.*;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Collection;
import java.util.Set;
import il.ac.tau.cs.dqsq.util.Logger;
import java.util.Map;

/**
 * @author pettelno
 *
 * A QSQ rewriting of a single Datalog rule. The QSQ rewriting is based
 * on the algorithm presented in the book "Foundations of databases", by
 * Abiteboul et al.
 *  
 */
class QsqRewriting
{
	private RewritableRule rule; // The original rule to rewrite.
	private String adornment;
	
	// List of Rule objects, each represents a supplementary rule, which
	// is the result of the QSQ rewriting.	
	private List supplementaryRules = new ArrayList();  
	
	// The output rule that defines the relation in the head of the original
	// rule from the last sup rule. For example, if the original rule
	// is r(x,y) :- s(x,z), t(z,y) then mainRule is r(x,y) :- sup2(x,y).
	private Rule mainRule; 
	
	private String supNamePrefix;
	
	QsqRewriting(RewritableRuleManager ruleManager, String adornment, Set intensionalRelations, Map extensionalRelations)
	{
    	// First check if the rule was already rewritten with this adornment.
    	if(ruleManager.isRewritten(adornment))
    	{
        	Logger.sendMessage("Rule " + ruleManager.getName() + " was already rewritten with adornment " + adornment + ".",
        					   Logger.LOG_LEVEL_DEBUG);
    		return;
    	}
		
		this.rule = ruleManager.getRule();
		this.adornment = adornment;
		this.rule.setAdornment(adornment, intensionalRelations);
		ruleManager.addAdornment(adornment);
		
		// Format of each sup relation name:
		// sup_<original_rule_name>_<adornment>_<index>.
		// The adornment is added to the sup relation name because a rule
		// may be rewritten several times, each time with a different
		// adornment (for example, see the RSG example in the book
		// "Foundations of databases", page 321). So adding the adornment
		// to the sup relation name ensures us that the name is unique.
		this.supNamePrefix = "sup_" + rule.getName() + "_" + adornment + "_";
		
		Logger.sendMessage("Adorned version of rule " + this.rule.getName() +
				     	   ": " + this.rule.toString(),
						   Logger.LOG_LEVEL_DEBUG);

		// The relations in the rule body are sorted by their
		// binding level in reverse order, so that the variables
		// get bound as early as possible to reduce the amount of
		// data to materialize. After sorting, the adornment of
		// the rule is set again.
		this.rule.sortRelationsByBindingLevel();
		this.rule.setAdornment(adornment, intensionalRelations);
		this.rule.removeExtraAdornments(extensionalRelations);
		
		Logger.sendMessage("Rule after sorting relation by binding level and after removing extra adornments: " +
					 	   this.rule.toString(),
						   Logger.LOG_LEVEL_DEBUG);

		setSupplementaryRules();
	}
		
	private void setSupplementaryRules()
	{
		setSup0();
		
		int numOfRelationsInBody = this.rule.getNumOfBodyRelations();
		for(int supIndex = 1; supIndex <= numOfRelationsInBody; supIndex++)
			setSupi(supIndex);
		
		setMainRule();
	}
	
	private void setSup0()
	{
		// Set rule head.
		String supHeadName = this.supNamePrefix + "0";
		List supVars = this.rule.getHeadBoundedVariables(this.adornment);
		RelationOccurrence supHead = new RelationOccurrence(
				supHeadName, this.rule.getHostingPeer(), this.rule.getHeadShortURL(),
				supVars, null);
		
		// Set rule body.
		String supBodyName = "in_" + this.rule.getHeadRelationName();
		RelationOccurrence supBodyRelation = new RelationOccurrence(
				supBodyName, this.rule.getHostingPeer(), this.rule.getHeadShortURL(),
				supVars, this.adornment);
		List supBody = new ArrayList();
		supBody.add(supBodyRelation);
		
		// Set rule.
		Rule rule = new Rule(supHead, supBody);
		this.supplementaryRules.add(rule);
		
		Logger.sendMessage("Set sup0 rule: " + rule.toString() + ".",
						   Logger.LOG_LEVEL_FLOW);		
	}

	private void setSupi(int index)
	{
		// Set rule head.
		String supHeadName = this.supNamePrefix + index;

		RelationOccurrence rel = this.rule.getRelationFromBody(index - 1);
		
		// The sup_i rule is hosted by the peer that hosts relation Ai in
		// the original rule, according to the terminology in the book.
		// (Relation Ai is the i'th relation in the original rule's body).
		String supPeerName = rel.getPeerName();
		String supURL = rel.getServiceShortURL();

		List supVars = setSupVars(index);

		RelationOccurrence supHead = new RelationOccurrence(
				supHeadName, supPeerName, supURL, supVars, null);

		// Set rule body. The body contains the previous sup relation
		// and relation Ai.
		List supBody = new ArrayList();
		RelationOccurrence prevSupRelation = ((Rule)this.supplementaryRules.get(index - 1)).getHeadRelation();
		supBody.add(prevSupRelation);
		supBody.add(this.rule.getRelationFromBody(index - 1));
		
		// Set rule.
		Rule rule = new Rule(supHead, supBody);
		this.supplementaryRules.add(rule);
		
		Logger.sendMessage("Set sup" + index + " rule: " + rule.toString() + ".",
						   Logger.LOG_LEVEL_FLOW);		
	}

	private List setSupVars(int index)
	{
		// According to the algorithm in the book, page 319:
		// For sup_n (the last sup) the variable set is the set of variables
		// For sup_i, 0<i<n, the variable set is the set Xi of variables
		// that occur both "before" Xi (i.e., occur in X0,A1,...,Ai)
		// and "after" Xi (i.e., occur in Ai+1,...,An,Xn).
		if(isLastSup(index))
			return this.rule.getHeadVariables(); 
		
		List supVars = new ArrayList();
		
		HashSet varsBeforeSup = getVarsBeforeSup(index);
		HashSet varsAfterSup = getVarsAfterSup(index);
		Iterator iter = varsBeforeSup.iterator();
		while(iter.hasNext())
		{
			String var = (String)iter.next();
			if(varsAfterSup.contains(var))
				supVars.add(var);
		}
		
		return supVars;
	}
	
	private HashSet getVarsBeforeSup(int supIndex)
	{
		// Initialize the set with the variables in sup_0 (Xi).
		Collection sup0Vars = this.getSup0().getHeadVariables();
		HashSet varsBeforeSup = new HashSet(sup0Vars);
		
		// Add the variables in A1,...Ai.
		for(int index = 1; index <= supIndex; index++)
			varsBeforeSup.addAll(this.rule.getBodyVariablesAtIndex(index - 1));
		
		return varsBeforeSup;
	}

	private HashSet getVarsAfterSup(int supIndex)
	{
		// Initialize the set with the variables in sup_n (Xn).
		// Note: Although sup_n was not set yet, we know that the variables
		// in it are all the variables in the original rule head.
		Collection supnVars = this.rule.getHeadVariables();
		HashSet varsAfterSup = new HashSet(supnVars);
		
		// Add the variables in Ai+1,...An.
		int numOfBodyRelations = this.rule.getNumOfBodyRelations();
		for(int index = supIndex + 1; index <= numOfBodyRelations; index++)
			varsAfterSup.addAll(this.rule.getBodyVariablesAtIndex(index - 1));
		
		return varsAfterSup;		
	}
	
	private void setMainRule()
	{
		// Set rule head.
		RelationOccurrence supHeadRelation = rule.getHeadRelation();
			
		// Set rule body.
		RelationOccurrence supBodyRelation = getLastSupRelation();
		
		List supBody = new ArrayList();
		supBody.add(supBodyRelation);
		
		// Set rule.
		Rule rule = new Rule(supHeadRelation, supBody);
		this.mainRule = rule;
		
		Logger.sendMessage("Set main rule: " + rule.toString() + ".",
						   Logger.LOG_LEVEL_FLOW);				
	}
	
	Rule getSup0()
	{
		return (Rule)this.supplementaryRules.get(0);
	}
	
	Rule getSupi(int index)
	{
		return (Rule)this.supplementaryRules.get(index);
	}

	Rule getMainRule()
	{
		return this.mainRule;
	}
	
	int numOfSupRules()
	{
		return this.supplementaryRules.size();
	}
	
	private boolean isLastSup(int supIndex)
	{
		if(supIndex == this.rule.getNumOfBodyRelations())
			return true;
		
		return false;
	}
	
	private	RelationOccurrence getLastSupRelation()
	{
		int numOfBodyRelations = this.rule.getNumOfBodyRelations();
		return ((Rule)this.supplementaryRules.get(numOfBodyRelations)).getHeadRelation();
	}
	
	boolean isEmpty()
	{
		return (this.supplementaryRules.size() == 0);
	}
	
	Rule getOriginalRuleAfterSort()
	{
		// The original rule is sorted (i.e., the relations in the
		// rule body are sorted) by the binding level, in the
		// constructor. So when calling this method, this.rule is
		// already sorted,
		return this.rule;
	}

/*	
    String[] serializeRules()
    {
	    int numOfSupRules = this.supplementaryRules.size();
	    String[] serializedRules = new String[numOfSupRules + 1];
	    for(int ruleIndex = 0; ruleIndex < numOfSupRules; ruleIndex++)
	    	serializedRules[ruleIndex] = this.supplementaryRules.get(ruleIndex).toString();
	
	    if(this.mainRule != null)
	    	serializedRules[numOfSupRules] = this.mainRule.toString();
	    
	    return serializedRules;
    }
*/    
}
