package il.ac.tau.cs.dqsq.datalog2qsq;

import il.ac.tau.cs.dqsq.common.Relation;
import il.ac.tau.cs.dqsq.common.Rule;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import org.jdom.Element;
import il.ac.tau.cs.dqsq.qsq2axml.Qsq2AxmlTranslator;
import il.ac.tau.cs.dqsq.util.Logger;
import il.ac.tau.cs.dqsq.util.dQSQUtils;

/**
 * @author pettelno
 *
 * A dQSQ query.
 * 
 */
public class Query
{
	private String name = null;
	private String bodyRelationName = null;
	private String adornment = "";
	private List boundedValues = new ArrayList();
	private List freeVars = new ArrayList();

	// The rule that defines the query, e.g. q(y) :- r("1", y).
	private Rule queryRule = null;
	
	// An "intensional query" is a query which its body relation
	// is intensional. Otherwise, it's an "extensional query" which
	// is a base-relation query such as/ q@R(x) :- parent@R("Alice", x)
	// in the ancestor Datalog program). 
	private boolean isIntensionalQuery = true;

	static private int id = 0;
	
	
/*	
	// Old code that received the query as a part of the input.	
	Query(Element queryElem)
	{
		Element headRelationElem = queryElem.getChild("head").getChild("relation");
		this.name = headRelationElem.getAttributeValue("name");
		Element bodyRelationElem = queryElem.getChild("body").getChild("relation");
		this.bodyRelationName = bodyRelationElem.getAttributeValue("name");
		setAdornment(bodyRelationElem.getChild("parameters").getChildren());
		setQueryRule(queryElem);
	}
*/
	
	Query(String queryAsString, Map extensionalRelations, String peerName) throws Exception
	{
		id++;
		Qsq2AxmlTranslator.setQueryIndex("q" + peerName + id);
		
		this.queryRule = new Rule(queryAsString);
		
		validate(peerName);
			
		this.name = this.queryRule.getHeadRelationName() + "_" + Qsq2AxmlTranslator.getQueryIndex();
//		this.name = this.queryRule.getHeadRelationName();
		this.bodyRelationName = this.queryRule.getBodysRelationName(0);
		
		if(extensionalRelations.containsKey(this.bodyRelationName))
			this.isIntensionalQuery = false;

		setAdornment(this.queryRule.getBodyVariablesAtIndex(0));

		// For an intensional query, add the adornment to the query,
		// e.g. the query q(y) :- r("1", y) becomes q(y) :- r_bf("1", y).
		if(this.isIntensionalQuery)
			this.queryRule.setBodyRelationAdornment(0, this.adornment);		
	}

	private void setAdornment(List bodyRelationParams)
	{
		Element paramElem = null;
		this.freeVars.add(0, null);
		int varIndex = 1;
		
		int numOfParams = bodyRelationParams.size();
		for(int paramIndex = 0; paramIndex < numOfParams; paramIndex++)
		{
			String paramVal = (String)bodyRelationParams.get(paramIndex);

			if(dQSQUtils.isParamBounded(paramVal))
			{
				this.adornment += "b";
				this.boundedValues.add(dQSQUtils.trimParamVal(paramVal));
				this.freeVars.add(varIndex, null);
				
				// If the query is intensional, replace the value of
				// the bounded variable in the query (a constant) with
				// *ANY*. Consider the ancestor Datalog program. If the
				// query is q(x) :- parent("Lois", x) (extensional)
				// then we want the value "Lois" to appear in rule, so
				// that only the children of "Lois" will be returned.
				// However, if the query is q(x) :- ancestor("Lois", x)
				// (intensional), the ancestor_bf relation will be
				// assigned with relavant tuples which do not necessarily
				// start with "Lois".
//				if(this.isIntensionalQuery)
//					bodyRelationParams.set(paramIndex, "*ANY*");
			}
			else
			{
				this.freeVars.add(varIndex, paramVal);
				this.adornment += "f";
			}
			
			varIndex++;
		}
		
		Logger.sendMessage("Query adornment: " + this.adornment + ".",
						   Logger.LOG_LEVEL_DEBUG);        			
	}

	
/*	
	private void setAdornment(List bodyRelationParams)
	{
		Element paramElem = null;
		this.freeVars.add(0, null);
		int varIndex = 1;
		Iterator iter = bodyRelationParams.iterator();
		while(iter.hasNext())
		{
			Object param = iter.next();

			String paramVal = null;
			
			if(param instanceof Element)
			{
				paramElem = (Element)iter.next();
				paramVal = paramElem.getText();
			}
			else
				paramVal = (String)param;

			if(dQSQUtils.isParamBounded(paramVal))
			{
				this.adornment += "b";
				this.boundedValues.add(dQSQUtils.trimParamVal(paramVal));
				this.freeVars.add(varIndex, null);
			}
			else
			{
				this.freeVars.add(varIndex, paramVal);
				this.adornment += "f";
			}
			
			varIndex++;
		}
		
		Logger.sendMessage("Query adornment: " + this.adornment + ".",
						   Logger.LOG_LEVEL_DEBUG);        			
	}
*/
	
/*
 	// Old code that received the query as a part of the input.	
	private void setAdornment(List bodyRelationParams)
	{
		Element paramElem = null;
		this.freeVars.add(0, null);
		int varIndex = 1;
		Iterator iter = bodyRelationParams.iterator();
		while(iter.hasNext())
		{
			paramElem = (Element)iter.next();
			String paramVal = paramElem.getText();
			if(dQSQUtils.isParamBounded(paramVal))
			{
				this.adornment += "b";
				this.boundedValues.add(dQSQUtils.trimParamVal(paramVal));
				this.freeVars.add(varIndex, null);
			}
			else
			{
				this.freeVars.add(varIndex, paramVal);
				this.adornment += "f";
			}
			
			varIndex++;
		}
		
		Logger.debug("Query adornment: " + this.adornment + ".");        			
	}
*/
	
	Relation setAdornedRelation()
	{
		// The adorned relation of the query is the relation in the body
		// of the query with its adornment.
		// For example, the adorned relation of the query q(y) :- r("1", y)
		// is r_bf.
		
		return new Relation(this.bodyRelationName, this.adornment, this.adornment.length(), null);
	}
	
	Relation setInputRelation()
	{
		// For example, the input relation of the query q(y) :- r("1", y)
		// is in_r_bf.
		
		// Transform the bounded variables to a tuples element in the form:
		//  <t>
		//   <at1>1</at1>
		//  </t>
		Element tupleElem = new Element("t");
		
		for(int index = 0; index < this.boundedValues.size(); index++)
		{
			// Start the <at> index from 1 (at1, at2,...).
			int atIndex = index + 1;
			Element atElem = new Element("at" + atIndex);
			atElem.setText((String)this.boundedValues.get(index));
			tupleElem.addContent(atElem);
		}
		
		// The list of <t> elements always contain one element.
		List tupleElems = new ArrayList();
		tupleElems.add(tupleElem);
		
		return new Relation("in_" + this.bodyRelationName, this.adornment, this.boundedValues.size(), tupleElems);	
	}
	
	String getBodyRelationName()
	{
		return this.bodyRelationName;
	}
	
	String getAdornment()
	{
		return this.adornment;
	}
	
	String getName()
	{
		return this.name;
	}
	
	Rule getQueryRule()
	{
		return this.queryRule; 
	}

/*	
 	// Old code that received the query as a part of the input.	
	private void setQueryRule(Element queryElem)
	{
		this.queryRule = new Rule(queryElem);
		
		// Add the adornment to the query, e.g. the query q(y) :- r("1", y)
		// becomes q(y) :- r_bf("1", y).
		this.queryRule.setBodyRelationAdornment(0, this.adornment);
	}
*/
	
	private void validate(String peerName) throws Exception
	{
		if(!this.queryRule.getHostingPeer().equals(peerName))
		{
			Logger.sendMessage("Peer in query's head must be current peer.",
					   		   Logger.LOG_LEVEL_ERROR);        	
			throw new Exception("Peer in query's head must be current peer.");
		}
		
		// Limitation - The query must be issued on the peer that hosts
		// the relation in the query body. For example, A query such as
		// q@P(x) :- parent@R("Lois", x) cannot be issued on peer P if
		// the parent relation is located on peer R. 
		if(!this.queryRule.getBodysRelationPeerName(0).equals(peerName))
		{
			Logger.sendMessage("Query must be issued on the peer that hosts the relation in the query body.",
					   		   Logger.LOG_LEVEL_ERROR);        	
			throw new Exception("Query must be issued on the peer that hosts the relation in the query body.");
		}		
	}
}
