package il.ac.tau.cs.dqsq.qsq2axml;

import org.jdom.output.XMLOutputter;
import org.jdom.CDATA;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import java.util.*;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import il.ac.tau.cs.dqsq.common.*;
import il.ac.tau.cs.dqsq.util.dQSQUtils;
import il.ac.tau.cs.dqsq.util.FileUtils;

/**
 * A dQSQ peer, responsible to translate the peer's part in the dQSQ
 * program to AXML format.
 * 
 * The QSQ2AXML translator can work in 2 modes:
 * 1. Dynamic mode: AXML documents & services are added while AXML
 * is already running. In this mode we don't have to explicitly add
 * the document to the repository. It is done automatically when
 * adding it with DynamicAxmlContentManager.addDocument.
 * In this mode the peer uses DynamicRelationTranslator and DynamicRuleTranslator
 * classes.
 * 2. Static mode (for testing): All the AXML documents & services are generated
 * before AXML starts.
 * In this mode the peer uses StaticRelationTranslator and StaticRuleTranslator
 * classes.
 */
abstract class Peer
{
	protected String peerName = null;
	protected Map relations = new HashMap();
	protected Collection rules = new ArrayList();
//    protected Element generalSettingsElem = null;
    protected String outputDirName = null;
    protected IActivityFile activityFile = null;
    protected INeighborsFile neighborsFile = null;
    protected IRepository axmlRepository = null;
    private final String BACKUP_REPOSITORY_DIR_NAME = "dQSQ.backup";
    
    Peer(String outputDirName) throws IOException, JDOMException
    {
    	this.outputDirName = outputDirName;
    }

    Peer(String outputDirName, String peerName)
		throws IOException, JDOMException
    {
//    	this.generalSettingsElem = generalSettingsElem;
    	this.peerName = peerName;
    	this.outputDirName = outputDirName;
    	backupOriginalRepository();
    }
    
    void run() throws Exception
    {
        writeOutput();    	
    }
    

    private void writeOutput() throws IOException, JDOMException
    {
        writePeerDocs();
    	writeRelations();
        writeRules();
        this.axmlRepository.flush();
//        neighborsFile.flush();
    }

    private void writeRelations() throws JDOMException, IOException
    {
        Collection relationInstances = relations.values();
        Iterator relationsIter = relationInstances.iterator();
        while(relationsIter.hasNext())
        {
            RelationTranslator rel = (RelationTranslator)relationsIter.next();
            writeRelation(rel);
        }
    }

    private void writeRelation(RelationTranslator rel) throws JDOMException, IOException
    {
        rel.write(this.outputDirName);
        this.activityFile.addRelation(rel.getName());
        this.neighborsFile.addRelation(rel.getName(), null);
//        this.neighborsFile.flush();
    }
    
    private void writeRules() throws IOException, JDOMException
    {
        Iterator rulesIter = rules.iterator();
        while(rulesIter.hasNext())
        {
            RuleTranslator rule = (RuleTranslator)rulesIter.next();
            rule.write(this.outputDirName, neighborsFile);
        }
    }
    
    private void writePeerDocs() throws IOException
    {
    	writeCallbackService();
    	writeIsFinishedService();
    	writeGetParamService();    	
//    	this.neighborsFile = new StaticNeighborsFile(this.outputDirName, generalSettingsElem.getChildText("peerName"));
    }

	private void writeCallbackService() throws IOException
	{
        Element root = new Element("serviceDefinition");
        root.setAttribute("name", "CallbackService");
        root.setAttribute("type", "callback");

        Element definitionElem = new Element("definition");
        
        Element transformServiceElem = new Element("transformService");
        transformServiceElem.setAttribute("URL", "");
        transformServiceElem.setAttribute("methodName", "");
        transformServiceElem.setAttribute("nameSpace", "");
        definitionElem.addContent(transformServiceElem);

        Element integrationServiceElem = new Element("integratorService");
        integrationServiceElem.setAttribute("URL", "");
        integrationServiceElem.setAttribute("methodName", "");
        integrationServiceElem.setAttribute("nameSpace", "");
        definitionElem.addContent(integrationServiceElem);

        root.addContent(definitionElem);
        
        Document outDoc = new Document(root);

        // Serialize it into a file.
        String serviceName = "CallbackService"; 
        FileOutputStream out = new FileOutputStream(outputDirName + File.separator + serviceName + ".xml");
        XMLOutputter serializer = new XMLOutputter();
        serializer.output(outDoc, out);
        out.flush();
        out.close();
        
        this.axmlRepository.addService(serviceName);        
	}

	private void writeIsFinishedService() throws IOException
	{
        Element root = new Element("serviceDefinition");
        root.setAttribute("name", "ContIsFinished");
        root.setAttribute("type", "continuous");
	        
        Element params = new Element("parameters");
        params.addContent(0, (new Element("param")).setAttribute("name", "nodeName"));
        params.addContent(1, (new Element("param")).setAttribute("name", "servicePeer"));
        root.addContent(params);
	        
        Element definitionElem = new Element("definition");
	        
        Element baseServiceElem = new Element("baseService");
        baseServiceElem.setAttribute("URL", dQSQUtils.getPeerURL() + "/servlet/AxisServlet");
        baseServiceElem.setAttribute("methodName", "isFinished");
        baseServiceElem.setAttribute("nameSpace", "IdlenessManager");
        definitionElem.addContent(baseServiceElem);

        Element transformServiceElem = new Element("transformService");
        transformServiceElem.setAttribute("URL", "");
        transformServiceElem.setAttribute("methodName", "");
        transformServiceElem.setAttribute("nameSpace", "");
        definitionElem.addContent(transformServiceElem);

        Element integrationServiceElem = new Element("integratorService");
        integrationServiceElem.setAttribute("URL", "");
        integrationServiceElem.setAttribute("methodName", "");
        integrationServiceElem.setAttribute("nameSpace", "");
        definitionElem.addContent(integrationServiceElem);

        Element frequencyElem = new Element("frequency");
        frequencyElem.setAttribute("default", "60000");
        frequencyElem.setAttribute("max", "1000");
        definitionElem.addContent(frequencyElem);

        Element subscriptionElem = new Element("subscription");
        subscriptionElem.setAttribute("maxLifeTime", "8640000");
        subscriptionElem.setAttribute("maxSuspensionTime", "86400");
        subscriptionElem.setAttribute("maxClientsNo", "15000");
        definitionElem.addContent(subscriptionElem);
	        
        root.addContent(definitionElem);

        Document outDoc = new Document(root);

        // Serialize it into a file.
        String serviceName = "ContIsFinished"; 
        FileOutputStream out = new FileOutputStream(outputDirName + File.separator + serviceName + ".xml");
        XMLOutputter serializer = new XMLOutputter();
        serializer.output(outDoc, out);
        out.flush();
        out.close();
	        
        this.axmlRepository.addService(serviceName);        
	}
	
	private void writeGetParamService() throws IOException
	{
        Element root = new Element("serviceDefinition");
        root.setAttribute("type", "query");

        Element params = new Element("parameters");
        params.addContent(0, (new Element("param")).setAttribute("name", "param"));
        root.addContent(params);
        root.addContent(new Element("definition").addContent(new Element("query").addContent(new CDATA("param;"))));
               
        Document outDoc = new Document(root);

        // Serialize it into a file.
        String serviceName = "getParam"; 
        FileOutputStream out = new FileOutputStream(outputDirName + File.separator + serviceName + ".xml");
        XMLOutputter serializer = new XMLOutputter();
        serializer.output(outDoc, out);
        out.flush();
        out.close();
        
        this.axmlRepository.addService(serviceName);        
	}
	
	void translateRelation(Relation rel, boolean isLeader) throws Exception
	{
		RelationTranslator relTranslator = new RelationTranslator(
												this.axmlRepository,
												rel,
												this.peerName,
												this.outputDirName,
												isLeader);
        writeRelation(relTranslator);        
	}

	void translateRule(Rule rule) throws Exception
	{
		RuleTranslator ruleTranslator = new RuleTranslator(
												this.axmlRepository,
												rule,
												this.peerName,
												this.outputDirName);
		ruleTranslator.write(this.outputDirName, neighborsFile);
	}
/*
	void translateQuery(Query query) throws IOException, JDOMException
	{
		new QueryTranslator(this.axmlRepository, query,
							this.generalSettingsElem, this.outputDirName);
	}
*/
/*	
	void translateQuery(String queryName) throws IOException
	{
		writeGetQueryResultService(queryName);
	}
	
	private void writeGetQueryResultService(String queryName) throws IOException
	{
		// Create the "getQueryResult" service. The purpose of this service is
		// to display the query result nicely. It simply takes the result
		// from the query relation document, e.g. for the query
		// q(y) :- r("1", y), the "getQueryResult" service takes the relevant data 
		// from q.xml.
        Element root = new Element("serviceDefinition");
        root.setAttribute("type", "query");
        
        Element params = new Element("parameters");
        root.addContent(params);
        
        root.addContent(new Element("definition").addContent
        		(new Element("query").addContent(
        				new CDATA("select tuple from tuple in " + queryName + "//data/tuples/t;"))));

        Document outDoc = new Document(root);

        // Serialize it into a file.
        String serviceName = "getQueryResult"; 
        FileOutputStream out = new FileOutputStream(outputDirName + File.separator + serviceName + ".xml");
        XMLOutputter serializer = new XMLOutputter();
        serializer.output(outDoc, out);
        out.flush();
        out.close();
        
        this.axmlRepository.addService(serviceName);  		
	}
*/

	String getName()
	{
		return this.peerName;
	}
	
	void clear() throws IOException
	{
    	// Remove and unregister all the AXML documents and services
    	// that were generated as a result of the computation of the
    	// current query.
		this.axmlRepository.clear();
    	restoreOriginalRepository();
	}
	
	void backupOriginalRepository() throws IOException
	{
		// Since we want to start each query with a clean repository, we
		// backup the original repository and will restore it when a new
		// query starts.
		File backupDir = new File(this.outputDirName + File.separatorChar + this.BACKUP_REPOSITORY_DIR_NAME);
		if(backupDir.exists())
			FileUtils.deleteFilesFromDir(backupDir);
		else
			backupDir.mkdir();
		
		File repositoryDir = new File(this.outputDirName);
		FileUtils.copyDir(repositoryDir, backupDir);
	}

	private void restoreOriginalRepository() throws IOException
	{
		File repositoryDir = new File(this.outputDirName);
		File backupDir = new File(this.outputDirName + File.separatorChar + this.BACKUP_REPOSITORY_DIR_NAME);
		FileUtils.deleteFilesFromDir(repositoryDir);
		FileUtils.copyDir(backupDir, repositoryDir);		
	}	
}
