package il.ac.tau.cs.dqsq.qsq2axml;

import java.io.IOException;
import org.jdom.JDOMException;
import il.ac.tau.cs.dqsq.common.*;
import il.ac.tau.cs.dqsq.util.Logger;
import java.util.Set;
import java.util.HashSet;

/**
 * @author pettelno
 *
 * The entry point to the QSQ2AXML module when running a full dQSQ
 * rewriting. It is used by the Datalog2QSQ module to translate rules
 * and relations to AXML format dynamically, meaning while the AXML
 * peer is already running.
 * 
 */
public class Qsq2AxmlTranslator
{
	private Peer peer = null;
	private Set translatedRelations = new HashSet();
	private int ruleIndex = 0; 

	// The dQSQ system can handle several queries sequentially. A new
	// query can be posed after the previous one had ended. We keep
	// track of the query index, due to a limitation in AXML. Although
	// all the AXML services and documents are unregistered when a 
	// query ends, we fail to add new documents or services with the
	// same name as the previous. Thus, as a workaround, we add the
	// query index to the name of the documents and services.
	// Query index format: <peer>_<ID>, where peer is the name of the
	// peer who started the query, and ID is the query counter on that
	// peer.
	private static String queryIndex; 
	
	public Qsq2AxmlTranslator(String outputDirName, String peerName)
		throws IOException, JDOMException
	{
		this.peer = new DynamicPeer(outputDirName, peerName);
		
		try
		{
			this.peer.run();
		}
		catch(Exception e)
		{
			Logger.sendMessage("Failed to run peer.", Logger.LOG_LEVEL_ERROR);        	
		}		
	}
	
	public void translateRelation(Relation rel, boolean isLeader)
	{
		// Translate relation unless it was already translated before.
		if(this.translatedRelations.contains(rel.getName()))
		{
			Logger.sendMessage("Relation " + rel.getName() + " was already translated in peer " + this.peer.getName() + ".",
							   Logger.LOG_LEVEL_FLOW);        	
			return;
		}
		
		Logger.sendMessage("Translating relation " + rel.getName() + ".",
					  	   Logger.LOG_LEVEL_DEBUG);        	
		
		try
		{
			this.peer.translateRelation(rel, isLeader);
			this.translatedRelations.add(rel.getName());
		}
		catch(Exception e)
		{
			Logger.sendMessage("Failed to translate relation " + rel.getName() + ".",
							   Logger.LOG_LEVEL_ERROR);        	
		}
	}
	
	public void translateRule(Rule rule)
	{		
		this.ruleIndex++;
		rule.setName(getQueryIndex() + "_qsq" + this.ruleIndex);
//		rule.setName("qsq" + this.ruleIndex);

		Logger.sendMessage("Translating rule " + 
					 	   rule.getName() + ": " + rule.toString() + ".",
						   Logger.LOG_LEVEL_DEBUG);        	
		
		try
		{
			this.peer.translateRule(rule);
		}
		catch(Exception e)
		{
			Logger.sendMessage("Failed to translate rule " +
						 	   rule.getName() + ": " + rule.toString() + " (" + e.getMessage() + ").",
							   Logger.LOG_LEVEL_ERROR);        	
		}		
	}
/*
	public void translateQuery(Query query)
	{
		Logger.debug("Translating query.");
	
		try
		{
			this.peer.translateQuery(query);
		}
		catch(Exception e)
		{
			Logger.error("Failed to translate query (" + e.getMessage() + ").");        	
		}		
	}
*/

/*	
	public void translateQueryService(String queryName)
	{
		Logger.sendMessage("Translating query " + queryName + ".",
						   Logger.LOG_LEVEL_FLOW);
		
		try
		{
			this.peer.translateQuery(queryName);
		}
		catch(Exception e)
		{
			Logger.sendMessage("Failed to translate query " + queryName + " (" + e.getMessage() + ").",
							   Logger.LOG_LEVEL_ERROR);        	
		}		
	}
*/
	
/*	
    public void refreshState(Element generalSettingsElem)
    {
    	refreshNeighborsMap(generalSettingsElem);
    }
*/

/*
    private void refreshNeighborsMap(Element generalSettingsElem)
    {
    	// The neighbors map for the termination detection algorithm is
    	// built dynamically, as the dQSQ rewriting is evolving. Here
    	// we call a web service to refresh the neighbors map of the
    	// only IdlenessManager object in this peer. Note that although
    	// it's on the same peer, we must use a web service because
    	// we need to update the same IdlenessManager object that is
    	// invoked when the isFinished web service is called from the
    	// AXML documents.
    	String localUrl = generalSettingsElem.getChildText("URL");
    	IdlenessManagerServiceLocator locator = new IdlenessManagerServiceLocator();
		try
		{
			URL serviceURL = new URL(localUrl + "/services/IdlenessManager");
			localhost.axml.services.IdlenessManager.IdlenessManager service = locator.getIdlenessManager(serviceURL);
			service.refreshNeighborsMap(this.peer.getName());
		}
		catch(Exception e)
		{
			Logger.error("Failed to refresh neighbors map");
		}    	
    }
*/
	
    public void clearPeer()
    {
    	// Remove and unregister all the AXML documents and services
    	// that were generated as a result of the computation of the
    	// current query.
		try
		{
//			queryIndex++;
			this.ruleIndex = 0;
			this.translatedRelations.clear();
			this.peer.clear();
			this.peer.run();
		}
		catch(Exception e)
		{
			Logger.sendMessage("Failed to clear peer.", Logger.LOG_LEVEL_ERROR);
		}    			
    }

/*    
    public void preQuery()
    {
    	try
		{
    		this.peer.backupRepository();
		}
		catch(Exception e)
		{
			Logger.error("Failed to backup repository.");
		}    			    	
    }
*/    
    
    public static String getQueryIndex()
    {
    	return queryIndex; 
    }
  
    public static void setQueryIndex(String qryIndex)
    {
    	queryIndex = qryIndex;
    }   
}
