package il.ac.tau.cs.dqsq.qsq2axml;

import org.jdom.Element;
import org.jdom.Document;
import org.jdom.Namespace;
import org.jdom.CDATA;
import org.jdom.output.XMLOutputter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.File;
import il.ac.tau.cs.dqsq.common.Relation;
import il.ac.tau.cs.dqsq.util.dQSQUtils;

/**
 * @author pettelno
 *
 * The translator of a dQSQ relation to AXML format. It builds the AXML
 * documents and services which are required for the relation in AXML:
 * 1. The main relation document. (For example, r.xml for relation r(x,y)).
 * 2. The relation base service which serves clients that want to get
 *    data from the relation (Service_r.xml).   
 * 3. The relation continuous service (ContService_r.xml).   
 * 4. The clients document of the relation continuous service, that keeps
 *    the list of all the registering clients, and for each one the history
 *    of data that was sent (ContService_r_clients.xml).
 * 
 */
class RelationTranslator
{
	private static final String TERMINATION_DETECTION_FREQUENCY = "10000";
	
	private Relation relation = new Relation();
	private String name;
    private int numOfAttributes;	
    private String outputDirName;
    private String localShortURL = null;
    private String localURL = null;
    private String localPeerName = null;
    private String serviceCallFrequency = null;
    private Namespace axmlNamespace = Namespace.getNamespace("axml", "http://www-rocq.inria.fr/verso/AXML");
    private IRepository axmlRepository = null;
    private boolean isLeader = false;
    
    RelationTranslator(IRepository axmlRepository, String peerName)
		throws Exception
    {
    	setGeneralSettings(peerName);
    	this.axmlRepository = axmlRepository;
    }


    RelationTranslator(IRepository axmlRepository, Relation rel, String peerName,
    		String outputDirName, boolean isLeader) throws Exception
    {
    	setGeneralSettings(peerName);
    	this.relation = rel;
    	this.outputDirName = outputDirName;
    	this.name = this.relation.getName();
    	this.numOfAttributes = this.relation.getnumOfAttributes();
    	this.axmlRepository = axmlRepository;
    	this.isLeader = isLeader;
    }
    
    private void setGeneralSettings(String peerName) throws Exception
    {
    	this.localShortURL = dQSQUtils.getPeerURL();
    	this.localURL = localShortURL + "/servlet/AxisServlet";
    	this.localPeerName = peerName;
    	this.serviceCallFrequency = dQSQUtils.getServiceCallFrequency();    	
    }
    
    void read(Element relationElem)
    {
    	this.relation.read(relationElem);
    	this.name = this.relation.getName();
    	this.numOfAttributes = this.relation.getnumOfAttributes();
    	String leaderAttrib = relationElem.getAttributeValue("leader");
    	if(leaderAttrib != null && leaderAttrib.equals("y"))
    		this.isLeader = true;
    }

    void write(String outputDirName) throws IOException
    {
        this.outputDirName = outputDirName;
        writeMainDocument();
        writeService();
        writeContService();        
        writeServiceClientsDoc();
        writeUpdateService();
    }

    private void writeMainDocument() throws IOException
    {
        Namespace axmlNS = Namespace.getNamespace("axml", "http://www-rocq.inria.fr/verso/AXML");
        Element root = new Element(this.name);
        root.setAttribute("docName", name, axmlNS);
        root.addContent(new Element("data").
        		addContent(new Element("tuples").addContent(this.relation.getTuples())));

        // If this relation is the leader, add the <termination> element
        // for the termination detection algorithm.
        if(this.isLeader)
        	addTerminationElement(root);

        // Another pathetic workaround:
        // Add a dummy service call to the main document.
        // Without it, <Rule> elements cannot be added to the document
        // if the relation appears in the head of a rule
        // (RuleTranslator.updateHeadsMainDocument).
        addDummyServiceCall(root);
        
        Document outDoc = new Document(root);

        // Serialize it into a file.
        FileOutputStream out = new FileOutputStream(outputDirName + File.separator + name + ".xml");
        XMLOutputter serializer = new XMLOutputter();
        serializer.output(outDoc, out);
        out.flush();
        out.close();
        
        this.axmlRepository.addDocument(name);
    }

    private void writeService() throws IOException
    {
        Element root = new Element("serviceDefinition");
        root.setAttribute("type", "query");
        
        Element params = new Element("parameters");
        params.addContent(0, (new Element("param")).setAttribute("name", "clientID"));
        params.addContent(1, (new Element("param")).setAttribute("name", "request"));
        params.addContent(2, (new Element("param")).setAttribute("name", "servicePeer"));
        params.addContent(3, (new Element("param")).setAttribute("name", "buildQueryService"));
        params.addContent(4, (new Element("param")).setAttribute("name", "buildQueryServiceURL"));
        params.addContent(5, (new Element("param")).setAttribute("name", "nextService"));
        params.addContent(6, (new Element("param")).setAttribute("name", "nextServiceURL"));
        params.addContent(7, (new Element("param")).setAttribute("name", "nextServicePeer"));
        params.addContent(8, (new Element("param")).setAttribute("name", "clientURL"));
        params.addContent(9, (new Element("param")).setAttribute("name", "clientRelationName"));
        params.addContent(10, (new Element("param")).setAttribute("name", "clientPeer"));
        params.addContent(11, (new Element("param")).setAttribute("name", "ruleName"));
        root.addContent(params);
        
        root.addContent(new Element("definition").addContent(new Element("query").addContent(new CDATA(generateCDATA()))));
        
        Document outDoc = new Document(root);

        // Serialize it into a file.
        String serviceName = "Service_" + name; 
        FileOutputStream out = new FileOutputStream(outputDirName + File.separator + serviceName + ".xml");
        XMLOutputter serializer = new XMLOutputter();
        serializer.output(outDoc, out);
        out.flush();
        out.close();
        
        this.axmlRepository.addService(serviceName);
    }

    private void writeContService() throws IOException
    {
        Element root = new Element("serviceDefinition");
        root.setAttribute("name", "ContService_" + this.name);
        root.setAttribute("type", "continuous");
        
        Element params = new Element("parameters");
        params.addContent(0, (new Element("param")).setAttribute("name", "clientID"));
        params.addContent(1, (new Element("param")).setAttribute("name", "request"));
        params.addContent(2, (new Element("param")).setAttribute("name", "servicePeer"));
        params.addContent(3, (new Element("param")).setAttribute("name", "buildQueryService"));
        params.addContent(4, (new Element("param")).setAttribute("name", "buildQueryServiceURL"));
        params.addContent(5, (new Element("param")).setAttribute("name", "nextService"));
        params.addContent(6, (new Element("param")).setAttribute("name", "nextServiceURL"));
        params.addContent(7, (new Element("param")).setAttribute("name", "nextServicePeer"));
        params.addContent(8, (new Element("param")).setAttribute("name", "clientURL"));
        params.addContent(9, (new Element("param")).setAttribute("name", "clientRelationName"));
        params.addContent(10, (new Element("param")).setAttribute("name", "clientPeer"));
        params.addContent(11, (new Element("param")).setAttribute("name", "ruleName"));
        root.addContent(params);
        
        Element definitionElem = new Element("definition");
        
        Element baseServiceElem = new Element("baseService");
        baseServiceElem.setAttribute("URL", this.localURL);
        baseServiceElem.setAttribute("methodName", "Service_" + this.name);
        baseServiceElem.setAttribute("nameSpace", "Service_" + this.name);
        definitionElem.addContent(baseServiceElem);

        Element transformServiceElem = new Element("transformService");
        transformServiceElem.setAttribute("URL", this.localURL);
        transformServiceElem.setAttribute("methodName", "doDiff");
        transformServiceElem.setAttribute("nameSpace", "DiffService");
        definitionElem.addContent(transformServiceElem);

        Element integrationServiceElem = new Element("integratorService");
        integrationServiceElem.setAttribute("URL", "");
        integrationServiceElem.setAttribute("methodName", "");
        integrationServiceElem.setAttribute("nameSpace", "");
        definitionElem.addContent(integrationServiceElem);

        Element frequencyElem = new Element("frequency");
        frequencyElem.setAttribute("default", "60000");
        frequencyElem.setAttribute("max", "1000");
        definitionElem.addContent(frequencyElem);

        Element subscriptionElem = new Element("subscription");
        subscriptionElem.setAttribute("maxLifeTime", "8640000");
        subscriptionElem.setAttribute("maxSuspensionTime", "86400");
        subscriptionElem.setAttribute("maxClientsNo", "15000");
        definitionElem.addContent(subscriptionElem);
        
        root.addContent(definitionElem);

        Document outDoc = new Document(root);

        // Serialize it into a file.
        String serviceName = "ContService_" + name; 
        FileOutputStream out = new FileOutputStream(outputDirName + File.separator + serviceName + ".xml");
        XMLOutputter serializer = new XMLOutputter();
        serializer.output(outDoc, out);
        out.flush();
        out.close();
        
        this.axmlRepository.addService(serviceName);
    }

	private void writeServiceClientsDoc() throws IOException
	{
	    Element root = new Element("Service_" + this.name + "_clients");
	    root.setAttribute("docName", "Service_" + this.name + "_clients", axmlNamespace);
	
	    Document outDoc = new Document(root);
	
	    // Serialize it into a file.
	    String docName = "ContService_" + this.name + "_clients"; 
	    FileOutputStream out = new FileOutputStream(outputDirName + File.separator + docName + ".xml");
	    XMLOutputter serializer = new XMLOutputter();
	    serializer.output(outDoc, out);
	    out.flush();
	    out.close();       
	}
    
    private String generateCDATA()
    {
        String cdataPart1 = "<reply>\n" +
							" <serviceName>ContService_" + this.name + "</>,\n" + 
							" <\"servicePeer\">servicePeer</>,\n" + 
							" <\"clientID\">clientID</>,\n" +
							" <\"clientRelationName\">clientRelationName</>,\n" +
							" <tuples>\n" +
							"  select\n" +
                            "   <t>\n";

        String cdataPart2 = "";
        for (int i = 1; i <= numOfAttributes; i++)
            cdataPart2 += "    <var" + i + ">" + "x/at" + i + "/text()</>,\n";

        String cdataPart3 = "\n    select\n" +
                              "     <extra_var_name>extra_var_val</>\n" +
                              "    from\n" +
                              "     extra_var in y/extra/var,\n" +
                              "     extra_var_name in extra_var/name,\n" +
                              "     extra_var_val in extra_var/value/text()\n" +
                              "   </>\n" +
//                              "  from x in " + name + "//t,\n";
        					  "  from x in " + name + "//data/tuples/t,\n";

        String cdataPart4 = "";
        for (int i = 1; i <= numOfAttributes; i++)
            cdataPart4 += "       var" + i + " in request/keys/at" + i + ",\n";

        String cdataPart5 = "       y in request/values/t,\n";

        String cdataPart6 = "";
        for (int i = 1; i < numOfAttributes; i++)
            cdataPart6 += "       val" + i + " in y/required/at" + i + ",\n";
        // no comma at the end of the line for last variable.
        cdataPart6 += "       val" + numOfAttributes + " in y/required/at" + numOfAttributes + "\n";

        String cdataPart7 = "  where ";
        for (int i = 1; i < numOfAttributes; i++)
            cdataPart7 +=    " (x/at" + i + "=val" + i + " or val" + i + "/text()=\"*ANY*\") and\n";
        cdataPart7 += "        (x/at" + numOfAttributes + "=val" + numOfAttributes + " or val" + numOfAttributes + "/text()=\"*ANY*\")\n";

        String cdataPart8 = " </>,\n\n";

        String cdataPart9 = " <services>\n\n" +
							" (select\n" +
							
							"  <incrementActivity>\n" +
							"   <sc xmlns=\"http://www-rocq.inria.fr/verso/AXML\" frequency=\"once\"\n" +
							"       serviceNameSpace=\"ActivityService\"\n" +
							"       methodName=\"incrementActivity\"\n" +
							"       serviceURL=concat(text(clientURL), \"/servlet/AxisServlet\")\n" +
							"       followedBy=\"../../Res_buildQuery/axml:sc\">\n" +
							"    <params xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"relationName\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">clientRelationName</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"cause\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"rule 4, after receiving data from first relation " + this.name + "\"</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"peer\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">clientPeer</>\n" +
							"     </>\n" +
							"    </>\n" +
							"   </>\n" +
							"  </>,\n\n" +
							
							"  <updateIdleness>\n" +
							"   <sc xmlns=\"http://www-rocq.inria.fr/verso/AXML\" frequency=\"once\"\n" +
							"       serviceNameSpace=\"IdlenessManager\"\n" +
							"       methodName=\"onRecvMsg\"\n" +
							"       serviceURL=concat(text(clientURL), \"/servlet/AxisServlet\")>\n" +
							"    <params xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"relationName\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">clientRelationName</>\n" +
							"     </>\n" +
							"    </>\n" +
							"   </>\n" +
							"  </>,\n\n" +
							
							" <Res_buildQuery>\n" +
							"  <sc xmlns=\"http://www-rocq.inria.fr/verso/AXML\" frequency=\"lazy\"\n" +
							"      methodName=buildQuery serviceNameSpace=buildQuery\n" +
							"      serviceURL=buildQueryServiceURL\n" +
							"      followedBy=\"../../Res_nextRelation/axml:sc[@serviceNameSpace='ActivityService']\">\n" +
							"    <params xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"param\">\n" +
							"      <xpath xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"../../../tuples\"</>\n" +
							"     </>\n" +
							"    </>\n" +
							"   </>\n" +
							"  </>,\n\n" +
							
							" <Res_nextRelation>\n\n" +
							
							"  <sc xmlns=\"http://www-rocq.inria.fr/verso/AXML\" frequency=\"lazy\"\n" +
							"      methodName=\"incrementActivity\" serviceNameSpace=\"ActivityService\"\n" +
							"      serviceURL=concat(text(clientURL), \"/servlet/AxisServlet\")\n" +
							"      followedBy=\"../axml:sc[@serviceNameSpace='RegisterService']\">\n" +
							"    <params xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"relationName\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">clientRelationName</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"cause\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">concat(\"rule 1, before registering on service \", text(nextService))</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"peer\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">clientPeer</>\n" +
							"     </>\n" +
							"    </>\n" +
							"   </>,\n\n" +
							
							"  <sc xmlns=\"http://www-rocq.inria.fr/verso/AXML\" frequency=\"lazy\"\n" +
							"      methodName=\"register\" serviceNameSpace=\"RegisterService\"\n" +
							"      serviceURL=nextServiceURL\n" +
							"      followedBy=\"../axml:sc[@subscription='true']\">\n" +
							"    <params xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"serviceName\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">nextService</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"clientID\">\n" +
							"      <xpath xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"./@id\"</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"clientURL\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">clientURL</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"clientRelationName\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">clientRelationName</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"servicePeer\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">nextServicePeer</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"clientPeer\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">clientPeer</>\n" +
							"     </>\n" +
							"    </>\n" +
							"   </>,\n\n" +
							
							"  <sc xmlns=\"http://www-rocq.inria.fr/verso/AXML\"\n" +
//							"      desiredFrequency=\"" + this.serviceCallFrequency + "\" frequency=\"lazy\"\n" +
							"      desiredFrequency=\"onEvent\" frequency=\"lazy\"\n" +
							"      methodName=nextServiceName serviceNameSpace=nextServiceName\n" +
							"      serviceURL=nextServiceURL subscription=\"true\" useDiff=\"true\"\n" +
							"      followedBy=\"../../ackService/axml:sc[@methodName='recvAck']\">\n" +
							"    <params xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"request\">\n" +
							"      <xpath xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"../../Res_buildQuery/request\"</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"clientID\">\n" +
							"      <xpath xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"../axml:sc[@serviceNameSpace='RegisterService']/@id\"</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"servicePeer\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">nextServicePeer</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"buildQueryService\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"*NONE*\"</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"buildQueryServiceURL\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"*NONE*\"</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"nextService\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"*NONE*\"</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"nextServiceURL\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"*NONE*\"</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"nextServicePeer\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"*NONE*\"</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"clientURL\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">clientURL</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"clientRelationName\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">clientRelationName</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"clientPeer\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">clientPeer</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"ruleName\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">ruleName</>\n" +
							"     </>\n" +
							"    </>\n" +
							"   </>\n" +
							"  </>,\n\n" +

        					" <ackService>\n" +
							"  <sc xmlns=\"http://www-rocq.inria.fr/verso/AXML\" frequency=\"lazy\"\n" +
							"      serviceNameSpace=\"AckService\"\n" +
							"      methodName=\"recvAck\"\n" +
							"      serviceURL=\"" + this.localURL + "\"\n" +
							"      followedBy=\"../axml:sc[@serviceNameSpace='ActivityService']\">\n" +
							"    <params xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"relationName\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"" + this.name + "\"</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"from\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">clientRelationName</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"receivingPeer\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"" + this.localPeerName + "\"</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"ackReason\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"after receiving data from first relation\"</>\n" +
							"     </>\n" +
							"    </>\n" +
							"  </>,\n" +

							"  <sc xmlns=\"http://www-rocq.inria.fr/verso/AXML\" frequency=\"lazy\"\n" +
							"      serviceNameSpace=\"ActivityService\"\n" +
							"      methodName=\"decrementActivity\"\n" +
							"      serviceURL=concat(text(clientURL), \"/servlet/AxisServlet\")>\n" +
							"    <params xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"relationName\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">clientRelationName</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"cause\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"rule 6, after sending ack on data to " + this.name +"\"</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"peer\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">clientPeer</>\n" +
							"     </>\n" +
							"    </>\n" +
							"   </>\n" +
							" </>\n\n" +

							" from buildQuery in buildQueryService,\n" +
							"      nextServiceName in nextService\n" +
							" where not(nextServiceName=\"*NONE*\"))\n\n" +

							" union\n" +
							" (select\n" +
/*
							"  <sc xmlns=\"http://www-rocq.inria.fr/verso/AXML\" frequency=\"once\"\n" +
							"      methodName=\"setShouldCheckData\" serviceNameSpace=\"RegisterService\"\n" +
							"      serviceURL=concat(text(clientURL), \"/servlet/AxisServlet\")>\n" +
							"    <params xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"relationName\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">ruleName</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"peer\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">clientPeer</>\n" +
							"     </>\n" +
							"    </>\n" +
							"  </>\n\n" +
*/
							"  <sc xmlns=\"http://www-rocq.inria.fr/verso/AXML\" frequency=\"once\"\n" +
							"      methodName=\"onReceiveData\" serviceNameSpace=\"GeneralService\"\n" +
							"      serviceURL=concat(text(clientURL), \"/servlet/AxisServlet\")>\n" +
							"    <params xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"receivingRelationName\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">ruleName</>\n" +
							"     </>\n" +
							"    </>\n" +
							"  </>\n\n" +
							
							"  from nextServiceName in nextService\n" +
							"  where nextServiceName=\"*NONE*\")\n\n" +
							" </>\n\n" +
							"</>;\n\n";
        
        return cdataPart1 + cdataPart2 + cdataPart3 + cdataPart4 + cdataPart5 + cdataPart6  + cdataPart7 + cdataPart8 + cdataPart9;
    }
    
    String getName()
    {
    	return this.name;
    }

    private void addTerminationElement(Element root)
    {
    	Element terminationElement = new Element("termination");

        // Add sc element.
//    	Element scElement = Qsq2AxmlUtils.createScElement(this.serviceCallFrequency, "once", "ContIsFinished", "ContIsFinished", 
//    			this.localURL, "replace", /* subscription */ "true", null, null); 
    	Element scElement = Qsq2AxmlUtils.createScElement(TERMINATION_DETECTION_FREQUENCY, "once", "ContIsFinished", "ContIsFinished", 
    			this.localURL, "replace", /* subscription */ "true", null, null); 

    	// Add params to service call.
    	Element paramsElem = new Element("params", axmlNamespace);

    	Element nodeNameParamElem = new Element("param", axmlNamespace).setAttribute("name", "nodeName");
    	nodeNameParamElem.addContent(new Element("value", axmlNamespace).setText(this.relation.getName()));
    	paramsElem.addContent(nodeNameParamElem);

    	Element servicePeerParamElem = new Element("param", axmlNamespace).setAttribute("name", "servicePeer");
    	servicePeerParamElem.addContent(new Element("value", axmlNamespace).setText(this.localPeerName));
    	paramsElem.addContent(servicePeerParamElem);

    	scElement.addContent(paramsElem);	
    	
    	terminationElement.addContent(scElement);
    	
    	root.addContent(terminationElement);   	    	
    }
    
    private void writeUpdateService() throws IOException
    {
        Element root = new Element("serviceDefinition");
        root.setAttribute("type", "update");
        
        Element params = new Element("parameters");
        params.addContent(0, (new Element("param")).setAttribute("name", "rule"));
        root.addContent(params);
        
        Element definitionElem = new Element("definition");
        Element actionElem = new Element("action").setAttribute("type", "insert");
        Element dataElem = new Element("data").addContent(new CDATA("rule;"));
        Element locationElem = new Element("location").addContent(
        		new CDATA("forest(\"" + this.getName() + "\")/\"" + 
        				  this.getName() + "\";"));
        actionElem.addContent(dataElem);
        actionElem.addContent(locationElem);
        definitionElem.addContent(actionElem);
        root.addContent(definitionElem);
                
        Document outDoc = new Document(root);

        // Serialize it into a file.
        String serviceName = "addRule_" + name; 
        FileOutputStream out = new FileOutputStream(outputDirName + File.separator + serviceName + ".xml");
        XMLOutputter serializer = new XMLOutputter();
        serializer.output(outDoc, out);
        out.flush();
        out.close();
        
        this.axmlRepository.addService(serviceName);
    }
    
    private void addDummyServiceCall(Element root)
    {
    	Element scElement = Qsq2AxmlUtils.createScElement(null, "once", "getParam", "getParam", 
    			this.localURL, "replace", null, null, null);
    	root.addContent(scElement);   	    	    	
    }
 }
