package il.ac.tau.cs.dqsq.qsq2axml;

import org.jdom.Element;
import org.jdom.CDATA;
import org.jdom.Document;
import org.jdom.Namespace;
import org.jdom.output.XMLOutputter;
import org.jdom.JDOMException;
import java.io.IOException;
import java.io.FileOutputStream;
import java.io.File;
import il.ac.tau.cs.dqsq.common.*;
import il.ac.tau.cs.dqsq.util.Logger;
import il.ac.tau.cs.dqsq.util.dQSQUtils;
import org.jdom.input.SAXBuilder;

/**
 * @author pettelno
 *
 * The translator of a dQSQ rule to AXML format. It builds the AXML
 * documents and services which are required for the rule in AXML:
 * 1. The buildQuery service which builds the query for the
 *    second relation in the rule body, based on the data returned from
 *    the first service. This service is created only if there are 2 
 *    relations in the rule body. (For example, buildQuery_qsq1.xml for
 *    rule qsq1).
 * 2. The help document which is an intermediate storage of the data
 *    returned from the relation in the rule, before it is sent to the 
 *    relaiton in the head of the rule. (Help_rule_qsq1.xml).
 * 3. The rule base service that used to send the relevant data from the
 *    help document to the relaiton in the head of the rule (Rule_qsq1.xml).
 * 4. The rule continuous service. (ContRule_qsq1.xml).
 * 5. The clients document of the rule continuous service, that keeps
 *    the list of all the registering clients (in this case there is always
 *    a single client - the relation in the head of the rule), and for each
 *    one the history of data that was sent (Rule_qsq1_clients.xml).
 * 
 */
class RuleTranslator
{
	private Rule rule = new Rule();
    private String name;
    private String outputDirName;
    private Namespace axmlNamespace = Namespace.getNamespace("axml", "http://www-rocq.inria.fr/verso/AXML");
    private String localShortURL = null;
    private String localURL = null;
    private String localPeerName = null;
    private String serviceCallFrequency = null;
    private IRepository axmlRepository = null;
        
    RuleTranslator(IRepository axmlRepository, String peerName)
		throws Exception
    {
    	setGeneralSettings(peerName);
    	this.axmlRepository = axmlRepository;    	
    }

    RuleTranslator(IRepository axmlRepository, Rule rule, String peerName, String outputDirName)
		throws Exception
	{
    	setGeneralSettings(peerName);
    	this.rule = rule;
    	this.outputDirName = outputDirName;
    	this.name = this.rule.getName();
    	this.axmlRepository = axmlRepository;    	
	}

    private void setGeneralSettings(String peerName) throws Exception
    {
    	this.localShortURL = dQSQUtils.getPeerURL();
    	this.localURL = localShortURL + "/servlet/AxisServlet";
    	this.localPeerName = peerName;
//    	this.serviceCallFrequency = dQSQUtils.getServiceCallFrequency();    	
    	this.serviceCallFrequency = "onEvent";    	
    }
    
    void read(Element ruleElem)
    {
    	this.rule.read(ruleElem);
    	this.name = this.rule.getName();
    }

    void write(String outputDirName, INeighborsFile neighborsFile) throws IOException, JDOMException
    {
        this.outputDirName = outputDirName;

        if(!this.rule.isSingleRelationInBody())
        	writeBuildQueryService();
		writeHelpDocument();
		writeRuleService();
		writeContRuleService();
		writeRuleClientsDoc();
		updateHeadsMainDocument();
		neighborsFile.addRelation(this.rule.getHeadRelationAdornedName(), this.rule.getBody());
//        neighborsFile.flush();
    }

    private void writeBuildQueryService() throws IOException
    {
        Element root = new Element("serviceDefinition");
        root.setAttribute("type", "query");
        root.addContent(new Element("parameters").addContent(new Element("param").setAttribute("name", "param")));
        root.addContent(new Element("definition").addContent(new Element("query").addContent(new CDATA(generateBuildQueryCDATA()))));

        Document outDoc = new Document(root);

        // Serialize it into a file.
        String serviceName = "buildQuery_" + name; 
        FileOutputStream out = new FileOutputStream(outputDirName + File.separator + serviceName + ".xml");
        XMLOutputter serializer = new XMLOutputter();
        serializer.output(outDoc, out);
        out.flush();
        out.close();
        
        this.axmlRepository.addService(serviceName);        
    }

    private String generateBuildQueryCDATA()
    {
        // Generate the input for the service of the target (second) relation in the
        // rule's body, based on the input of the call to the service of the source
        // (first) relation.
        RelationOccurrence srcRelation = this.rule.getRelationFromBody(0);
        int numOfAttributesInSrcRelation = srcRelation.numberOfAttributes();

        RelationOccurrence targetRelation = this.rule.getRelationFromBody(1);
        int numOfAttributesInTargetRelation = targetRelation.numberOfAttributes();

        String cdataPart1 = "<request>\n" +
                            " <keys>\n";

        String cdataPart2 = "";
        for (int i = 1; i < numOfAttributesInTargetRelation; i++)
            cdataPart2 += "   <at" + i + ">" + targetRelation.getAttributeName(i - 1) + "</>,\n";
        cdataPart2 += "   <at" + numOfAttributesInTargetRelation + ">" + targetRelation.getAttributeName(numOfAttributesInTargetRelation - 1) + "</>\n";

        String cdataPart3 = " </>,\n" +
                            " <values>\n" +
                            "  select\n" +
                            "   <t>\n" +
                            "    <required>\n";

        String cdataPart4 = "";
        for (int i = 1; i <= numOfAttributesInTargetRelation; i++)
        {
            // For each variable in the second relation, check if bound i.e., it
            // occurrs in the first relation.
            String curAttributeName = targetRelation.getAttributeName(i - 1);
            if(isBound(curAttributeName))
                cdataPart4 += "     <at" + i + ">" + "tuple/" + curAttributeName + "/text()</>";
            else
                cdataPart4 += "     <at" + i + ">" + "\"*ANY*\"</>";
            if(i < numOfAttributesInTargetRelation) // not last attribute
                cdataPart4 += ",";
            cdataPart4 += "\n";
        }

        String cdataPart5 = "    </>";
        boolean isFirstExtraVar = true;
        boolean areExtraVars = false;
        for (int i = 1; i <= numOfAttributesInSrcRelation; i++)
        {
            // For each variable in the first relation, check if it is extra i.e., it
            // does not occur in the second relation.
            String curAttributeName = srcRelation.getAttributeName(i - 1);
            if(isExtra(curAttributeName))
            {
                if(isFirstExtraVar)
                {
                    isFirstExtraVar = false;
                    areExtraVars = true;
                    cdataPart5 +=	",\n" +
                    				"    <extra>\n";
                }
                else
                	cdataPart5 += ",\n";               	
                
                cdataPart5 += "     <var>\n" +
                              "      <name>" + curAttributeName + "</>,\n" +
                              "      <value>tuple/" + curAttributeName + "/text()</>\n" +
                              "     </>";
                              
            }
        }

        cdataPart5 += "\n";
        
        if(areExtraVars)
        	cdataPart5 += "    </>\n";
        	
        String cdataPart6 = "   </>\n" +
        					"  from tuple in param/t\n" +
							" </>\n" +
							"</>;";
 
        
        return cdataPart1 + cdataPart2 + cdataPart3 + cdataPart4 + cdataPart5 + cdataPart6;
    }

    private boolean isBound(String varName)
    {
        return (this.rule.getRelationFromBody(0)).contains(varName);
    }

    private boolean isExtra(String varName)
    {
        return !(this.rule.getRelationFromBody(1)).contains(varName);
    }

    private void writeHelpDocument() throws IOException
    {
        String docName = "Help_rule_" + name;
    	Element root = new Element(docName);
        root.setAttribute("docName", docName, axmlNamespace);

        Element firstRelationElem = new Element("Res_" + this.rule.getRelationFromBody(0).getAdornedName());

        firstRelationElem.addContent(createIncActivityScElement());        
        firstRelationElem.addContent(createRegisterScElement());        
        firstRelationElem.addContent(createFirstRelationScElement());        

        root.addContent(firstRelationElem);
        
        Document outDoc = new Document(root);

        // Serialize it into a file.
        FileOutputStream out = new FileOutputStream(outputDirName + File.separator + docName + ".xml");
        XMLOutputter serializer = new XMLOutputter();
        serializer.output(outDoc, out);
        out.flush();
        out.close();
        
        this.axmlRepository.addDocument(docName);            
    }

    private Element createIncActivityScElement()
    {
    	Element scElem = Qsq2AxmlUtils.createScElement(null, "once", "incrementActivity", "ActivityService", this.localURL,
    			null, null, null, "../axml:sc[@serviceNameSpace='RegisterService']");

    	Element paramsElem = new Element("params", axmlNamespace);

    	Element relationNameParamElement = new Element("param", axmlNamespace).setAttribute("name", "relationName");
    	relationNameParamElement.addContent(new Element("value", axmlNamespace).setText(this.rule.getHeadRelationAdornedName()));
    	paramsElem.addContent(relationNameParamElement);

    	RelationOccurrence firstRelation = this.rule.getRelationFromBody(0);
    	Element causeParamElement = new Element("param", axmlNamespace).setAttribute("name", "cause");
    	causeParamElement.addContent(new Element("value", axmlNamespace).setText("rule 1, before registering on service ContService_" + firstRelation.getAdornedName()));
    	paramsElem.addContent(causeParamElement);

    	Element peerParamElement = new Element("param", axmlNamespace).setAttribute("name", "peer");
    	peerParamElement.addContent(new Element("value", axmlNamespace).setText(this.localPeerName));
    	paramsElem.addContent(peerParamElement);

    	scElem.addContent(paramsElem);

    	return scElem;
    }

    private Element createRegisterScElement()
    {
    	RelationOccurrence firstRelation = this.rule.getRelationFromBody(0);
    	
    	Element scElem = Qsq2AxmlUtils.createScElement(null, "lazy", "register", "RegisterService",
    			firstRelation.getServiceURL(), null, null, null, "../axml:sc[@serviceNameSpace='ContService_" + firstRelation.getAdornedName() + "']");

    	Element paramsElem = new Element("params", axmlNamespace);

    	Element serviceNameParamElement = new Element("param", axmlNamespace).setAttribute("name", "serviceName");
    	serviceNameParamElement.addContent(new Element("value", axmlNamespace).setText("ContService_" + firstRelation.getAdornedName()));
    	paramsElem.addContent(serviceNameParamElement);

    	Element clientIDParamElement = new Element("param", axmlNamespace).setAttribute("name", "clientID");
    	clientIDParamElement.addContent(new Element("xpath", axmlNamespace).setText("./@id"));
    	paramsElem.addContent(clientIDParamElement);

    	Element clientURLParamElement = new Element("param", axmlNamespace).setAttribute("name", "clientURL");
    	clientURLParamElement.addContent(new Element("value", axmlNamespace).setText(this.localShortURL));
    	paramsElem.addContent(clientURLParamElement);

    	Element clientRelationNameParamElement = new Element("param", axmlNamespace).setAttribute("name", "clientRelationName");
    	clientRelationNameParamElement.addContent(new Element("value", axmlNamespace).setText(this.rule.getHeadRelationAdornedName()));
    	paramsElem.addContent(clientRelationNameParamElement);

    	Element servicePeerParamElement = new Element("param", axmlNamespace).setAttribute("name", "servicePeer");
    	servicePeerParamElement.addContent(new Element("value", axmlNamespace).setText(firstRelation.getPeerName()));
    	paramsElem.addContent(servicePeerParamElement);

    	Element clientPeerParamElement = new Element("param", axmlNamespace).setAttribute("name", "clientPeer");
    	clientPeerParamElement.addContent(new Element("value", axmlNamespace).setText(this.localPeerName));
    	paramsElem.addContent(clientPeerParamElement);
    	
    	scElem.addContent(paramsElem);

    	return scElem;
    }

    private Element createFirstRelationScElement()
    {
    	RelationOccurrence firstRelation = this.rule.getRelationFromBody(0);
    	String firstRelationName = firstRelation.getAdornedName();

    	Element scElem = Qsq2AxmlUtils.createScElement(this.serviceCallFrequency, "lazy", "ContService_" + firstRelationName, "ContService_" + firstRelationName, 
    			firstRelation.getServiceURL(), null, "true", "true", null);
    	Element paramsElem = new Element("params", axmlNamespace);

    	Element clientIDParamElement = new Element("param", axmlNamespace).setAttribute("name", "clientID");
    	clientIDParamElement.addContent(new Element("xpath", axmlNamespace).setText("//axml:sc[@serviceNameSpace='RegisterService']/@id"));
    	paramsElem.addContent(clientIDParamElement);
    	   	
    	Element requestParamElement = new Element("param", axmlNamespace).setAttribute("name", "request");
    	Element valueElement = new Element("value", axmlNamespace);
    	Element requestElement = new Element("request");
    	
    	Element keysElement = new Element("keys");
    	for(int i = 1; i <= firstRelation.numberOfAttributes(); i++)
    	{
    		// If the rule is a query, e.g. q(y) :- r("1", y), then at 
    		// least one of the variables in the body relation is a 
    		// constant ("1" in the example). In such a case we set the
    		// variable in "keys" as "*DUMMY*" (it is meaningless).
    		String varName = firstRelation.getAttributeName(i - 1);
    		if(dQSQUtils.isParamBounded(varName))
    			varName = "*DUMMY*";
    		Element atElement = new Element("at" + i).setText(varName);
    		keysElement.addContent(atElement);
    	}

    	Element valuesElement = new Element("values");
    	Element tElement = new Element("t");
    	Element requiredElement = new Element("required");
    	for(int i = 1; i <= firstRelation.numberOfAttributes(); i++)
    	{
    		String requiredVarName = "*ANY*";
    		String origVarName = firstRelation.getAttributeName(i - 1);

    		// If the variable is bounded to a constant, we set the
    		// variable in "required" to the constant value instead of "*ANY*".
    		if(dQSQUtils.isParamBounded(origVarName))
    			requiredVarName = dQSQUtils.trimParamVal(origVarName);
    			
    		Element atElement = new Element("at" + i).setText(requiredVarName);
    		requiredElement.addContent(atElement);
    	}

		tElement.addContent(requiredElement);
		valuesElement.addContent(tElement);
		requestElement.addContent(keysElement);
		requestElement.addContent(valuesElement);
		valueElement.addContent(requestElement);
		requestParamElement.addContent(valueElement);
   		paramsElem.addContent(requestParamElement);
    	
    	Element servicePeerParamElement = new Element("param", axmlNamespace).setAttribute("name", "servicePeer");
    	servicePeerParamElement.addContent(new Element("value", axmlNamespace).setText(firstRelation.getPeerName()));
    	paramsElem.addContent(servicePeerParamElement);

    	// The buildQuery and nextService params get the value "*NONE*" if there
    	// is only one relation in the rule's body.
    	String buildQueryService = "*NONE*";
    	String buildQueryServiceURL = "*NONE*";
    	String nextService = "*NONE*";
    	String nextServiceURL = "*NONE*";
    	String nextServicePeer = "*NONE*";
    	if(!this.rule.isSingleRelationInBody())
    	{
	    	RelationOccurrence nextRelation = this.rule.getRelationFromBody(1);    		
        	buildQueryService = "buildQuery_" + this.name;
        	buildQueryServiceURL = this.localURL;
        	nextService = "ContService_" + nextRelation.getAdornedName();
        	nextServiceURL = nextRelation.getServiceURL();
        	nextServicePeer = nextRelation.getPeerName();    		
    	}

    	Element buildQueryParamElement = new Element("param", axmlNamespace).setAttribute("name", "buildQueryService");
	    buildQueryParamElement.addContent(new Element("value", axmlNamespace).setText(buildQueryService));
	    paramsElem.addContent(buildQueryParamElement);
	
	    Element buildQueryURLParamElement = new Element("param", axmlNamespace).setAttribute("name", "buildQueryServiceURL");
	    buildQueryURLParamElement.addContent(new Element("value", axmlNamespace).setText(buildQueryServiceURL));
	    paramsElem.addContent(buildQueryURLParamElement);
	    	
	    Element nextServiceParamElement = new Element("param", axmlNamespace).setAttribute("name", "nextService");
	    nextServiceParamElement.addContent(new Element("value", axmlNamespace).setText(nextService));
	    paramsElem.addContent(nextServiceParamElement);
	
	    Element nextServiceURLParamElement = new Element("param", axmlNamespace).setAttribute("name", "nextServiceURL");
	    nextServiceURLParamElement.addContent(new Element("value", axmlNamespace).setText(nextServiceURL));
	    paramsElem.addContent(nextServiceURLParamElement);
	
	    Element nextServicePeerParamElement = new Element("param", axmlNamespace).setAttribute("name", "nextServicePeer");
	    nextServicePeerParamElement.addContent(new Element("value", axmlNamespace).setText(nextServicePeer));
	    paramsElem.addContent(nextServicePeerParamElement);
    	
    	Element clientURLParamElement = new Element("param", axmlNamespace).setAttribute("name", "clientURL");
    	clientURLParamElement.addContent(new Element("value", axmlNamespace).setText(this.localShortURL));
    	paramsElem.addContent(clientURLParamElement);

    	Element clientRelationNameParamElement = new Element("param", axmlNamespace).setAttribute("name", "clientRelationName");
    	clientRelationNameParamElement.addContent(new Element("value", axmlNamespace).setText(this.rule.getHeadRelationAdornedName()));
    	paramsElem.addContent(clientRelationNameParamElement);

    	Element clientPeerParamElement = new Element("param", axmlNamespace).setAttribute("name", "clientPeer");
    	clientPeerParamElement.addContent(new Element("value", axmlNamespace).setText(this.localPeerName));
    	paramsElem.addContent(clientPeerParamElement);

    	Element ruleNameParamElement = new Element("param", axmlNamespace).setAttribute("name", "ruleName");
    	ruleNameParamElement.addContent(new Element("value", axmlNamespace).setText("Rule_" + this.name));
    	paramsElem.addContent(ruleNameParamElement);
    	
    	scElem.addContent(paramsElem);

    	return scElem;
    }

    private void writeRuleService() throws IOException
    {
        Element root = new Element("serviceDefinition");
        root.setAttribute("type", "query");
        
        Element params = new Element("parameters");
        params.addContent(0, (new Element("param")).setAttribute("name", "clientID"));
        params.addContent(1, (new Element("param")).setAttribute("name", "servicePeer"));
        root.addContent(params);
        
        root.addContent(new Element("definition").addContent(new Element("query").addContent(new CDATA(generateRuleServiceCDATA()))));

        Document outDoc = new Document(root);

        // Serialize it into a file.
        String serviceName = "Rule_" + name; 
        FileOutputStream out = new FileOutputStream(outputDirName + File.separator + serviceName + ".xml");
        XMLOutputter serializer = new XMLOutputter();
        serializer.output(outDoc, out);
        out.flush();
        out.close();
        
        this.axmlRepository.addService(serviceName);  
    }

    private void writeContRuleService() throws IOException
    {
        Element root = new Element("serviceDefinition");
        root.setAttribute("name", "ContRule_" + this.name);
        root.setAttribute("type", "continuous");
        
        Element params = new Element("parameters");
        params.addContent(0, (new Element("param")).setAttribute("name", "clientID"));
        params.addContent(1, (new Element("param")).setAttribute("name", "servicePeer"));
        root.addContent(params);
        
        Element definitionElem = new Element("definition");
        
        Element baseServiceElem = new Element("baseService");
        baseServiceElem.setAttribute("URL", this.localURL);
        baseServiceElem.setAttribute("methodName", "Rule_" + this.name);
        baseServiceElem.setAttribute("nameSpace", "Rule_" + this.name);
        definitionElem.addContent(baseServiceElem);

        Element transformServiceElem = new Element("transformService");
        transformServiceElem.setAttribute("URL", this.localURL);
        transformServiceElem.setAttribute("methodName", "doDiff");
        transformServiceElem.setAttribute("nameSpace", "DiffService");
        definitionElem.addContent(transformServiceElem);

        Element integrationServiceElem = new Element("integratorService");
        integrationServiceElem.setAttribute("URL", "");
        integrationServiceElem.setAttribute("methodName", "");
        integrationServiceElem.setAttribute("nameSpace", "");
        definitionElem.addContent(integrationServiceElem);

        Element frequencyElem = new Element("frequency");
        frequencyElem.setAttribute("default", "60000");
        frequencyElem.setAttribute("max", "1000");
        definitionElem.addContent(frequencyElem);

        Element subscriptionElem = new Element("subscription");
        subscriptionElem.setAttribute("maxLifeTime", "8640000");
        subscriptionElem.setAttribute("maxSuspensionTime", "86400");
        subscriptionElem.setAttribute("maxClientsNo", "15000");
        definitionElem.addContent(subscriptionElem);
        
        root.addContent(definitionElem);

        Document outDoc = new Document(root);

        // Serialize it into a file.
        String serviceName = "ContRule_" + this.name; 
        FileOutputStream out = new FileOutputStream(outputDirName + File.separator + serviceName + ".xml");
        XMLOutputter serializer = new XMLOutputter();
        serializer.output(outDoc, out);
        out.flush();
        out.close();
        
        this.axmlRepository.addService(serviceName);      	
    }

    private void writeRuleClientsDoc() throws IOException
    {
        Element root = new Element("Rule_" + this.name + "_clients");
        root.setAttribute("docName", "Rule_" + this.name + "_Clients", axmlNamespace);

        Document outDoc = new Document(root);

        // Serialize it into a file.
        String docName = "Rule_" + this.name + "_clients"; 
        FileOutputStream out = new FileOutputStream(outputDirName + File.separator + docName + ".xml");
        XMLOutputter serializer = new XMLOutputter();
        serializer.output(outDoc, out);
        out.flush();
        out.close();       
    }
    
    private String generateRuleServiceCDATA()
    {  	
        int numOfAttributesInHeadRelation = this.rule.getHeadsNumOfAttributes();

        String cdataPart1 = "<data>\n" +
							" <serviceName>" + "Rule_" + this.name + "</>,\n" +
							" <\"servicePeer\">servicePeer</>,\n" +
							" <\"clientID\">clientID</>,\n\n";
//							" <tuples/>,\n\n";
/*
    	String cdataPart2 = " <tuples>\n" +
    						"  select\n" +
                            "   <t>\n";

        String cdataPart3 = "";
        for (int i = 1; i < numOfAttributesInHeadRelation; i++)
            cdataPart3 += "    <at" + i + ">tuple/" + head.getAttributeName(i - 1) + "/text()</>,\n";
        cdataPart3 += "    <at" + numOfAttributesInHeadRelation + ">tuple/" + head.getAttributeName(numOfAttributesInHeadRelation - 1) + "/text()</>\n";

        String resElementInHelpDoc = "Res_nextRelation"; // Default, in case there are 2 relations in rule's body.
        if(isSingleRelationInBody()) // There is only one relation in rule's body.
        	resElementInHelpDoc = "Res_" + ((RelationOccurrence)body.get(0)).getName();
        
        String cdataPart4 = "   </>\n" +
                            "  from tuple in Help_rule_" + name + "//" + resElementInHelpDoc +"/reply/tuples/t\n" +
							" </>,\n";

        RelationOccurrence lastRelationInBody = null;
        if(isSingleRelationInBody())
			lastRelationInBody = (RelationOccurrence)this.body.get(0);
        else
			lastRelationInBody = (RelationOccurrence)this.body.get(1);
*/        
        
        
//		String cdataPart2 = " <services>\n" +
		String cdataPart2 =	"  <sc xmlns=\"http://www-rocq.inria.fr/verso/AXML\" frequency=\"once\"\n" +
							"      methodName=\"getParam\" serviceNameSpace=\"getParam\"\n" +
							"      serviceURL=\"" + this.localURL + "\"\n" +
							"      followedBy=\"../axml:sc[@methodName='postRecvDataFromLastService']\">\n" +

//					    	"    <merge-element-rule xmlns=\"http://www-rocq.inria.fr/verso/AXML\" mode=\"merge\">\n" +
//					    	"     <key xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\n" +
//							"      <selector xmlns=\"http://www-rocq.inria.fr/verso/AXML\" xpath=\"../..\"></>\n" +
//							"     </>\n" +
//							"    </>,\n" +

							"    <params xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"param\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\n";
							
	    String cdataPart3 = "       <tuples>\n" +
    						"        select\n" +
                            "        <t>\n";

        String cdataPart4 = "";
        for (int i = 1; i < numOfAttributesInHeadRelation; i++)
            cdataPart4 += "         <at" + i + ">tuple/" + this.rule.getHeadsAttributeName(i - 1) + "/text()</>,\n";
        cdataPart4 += "         <at" + numOfAttributesInHeadRelation + ">tuple/" + this.rule.getHeadsAttributeName(numOfAttributesInHeadRelation - 1) + "/text()</>\n";

        String resElementInHelpDoc = "Res_nextRelation"; // Default, in case there are 2 relations in rule's body.
        if(this.rule.isSingleRelationInBody()) // There is only one relation in rule's body.
        	resElementInHelpDoc = "Res_" + this.rule.getBodysRelationAdornedName(0);
        
        String cdataPart5 = "        </>\n" +
                            "       from tuple in Help_rule_" + name + "//" + resElementInHelpDoc +"/reply/tuples/t\n" +
							"      </>\n" +
							"     </>\n" +
							"    </>\n" +
							"   </>\n" +
        					"  </>,\n";


        RelationOccurrence lastRelationInBody = null;
        if(this.rule.isSingleRelationInBody())
			lastRelationInBody = this.rule.getRelationFromBody(0);
        else
			lastRelationInBody = this.rule.getRelationFromBody(1);
        
        String cdataPart6 =	"  <sc xmlns=\"http://www-rocq.inria.fr/verso/AXML\" frequency=\"lazy\"\n" +
							"      methodName=\"postRecvDataFromLastService\" serviceNameSpace=\"GeneralService\"\n" +
							"      serviceURL=\"" + this.localURL + "\">\n" +
							"    <params xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"serviceName\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"" + lastRelationInBody.getAdornedName() + "\"</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"clientRelationName\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"" + this.rule.getHeadRelationAdornedName() + "\"</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"clientURL\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"" + this.localShortURL + "\"</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"serverURL\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"" + lastRelationInBody.getServiceShortURL() + "\"</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"clientPeer\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"" + this.localPeerName + "\"</>\n" +
							"     </>,\n" +
							"     <param xmlns=\"http://www-rocq.inria.fr/verso/AXML\" name=\"serverPeer\">\n" +
							"      <value xmlns=\"http://www-rocq.inria.fr/verso/AXML\">\"" + lastRelationInBody.getPeerName() + "\"</>\n" +
							"     </>\n" +
							"    </>\n" +
							"  </>\n\n" +
//							" </>\n\n" +
							"</>;";
        
		return cdataPart1 + cdataPart2 + cdataPart3 + cdataPart4 + cdataPart5 + cdataPart6;
//        return cdataPart1 + cdataPart6;
    }

/*    
    private void updateHeadsMainDocument() throws IOException, JDOMException
    {
        SAXBuilder builder = new SAXBuilder();
        String docName = outputDirName + File.separator + this.rule.getHeadRelationAdornedName() + ".xml";
        Document doc = builder.build(new File(docName));
        Element root = doc.getRootElement();
    	
        // If there exists a data element, it needs to be located
        // after the service call elements. So remove it here...
//        Element dataElem = root.getChild("data");
//        if(dataElem != null)
//        	root.removeChild("data");
        
        addRuleElement(root);

//	    addRegisterServiceCall(root);
//	    addRuleServiceCall(root);
        
        // If the relation in the head of the rule is the leader,
        // add the <termination> element for the termination
        // detection algorithm.
//        if(this.rule.isHeadRelationLeader())
//        	addTerminationElement(root);
        
        // ... and add it here.
//        if(dataElem != null)
//        	root.addContent(dataElem);
        	
        // Update file.
        FileOutputStream out = new FileOutputStream(docName);
        XMLOutputter serializer = new XMLOutputter();
        serializer.output(doc, out);
        out.flush();
        out.close();
        
        this.axmlRepository.updateDocument(this.rule.getHeadRelationAdornedName());
    }
*/

    private void updateHeadsMainDocument() throws IOException, JDOMException
	{
    	// Yet another ugly workaround due to problems in AXML.
    	// We need to add a <Rule> element with a service call to the
    	// corresponding rule to the document of the relation in the
    	// head of the rule.
    	// Ideally, we should always use gentleUpdateHeadsMainDocument
    	// which adds the element by calling the addRule update service.
    	// However, it fails for the first rule. So the first rule is
    	// added roughly by physically writing to the document.
		Logger.sendMessage("RuleTranslator.updateHeadsMainDocument started", Logger.LOG_LEVEL_FLOW);        	

//		if(!roughUpdateHeadsMainDocument())
			gentleUpdateHeadsMainDocument();
		
/*		
    	if(isFirstRuleInHeadsMainDocument())
    	{
    		roughUpdateHeadsMainDocument();
    		Logger.sendMessage("RuleTranslator.updateHeadsMainDocument after roughUpdateHeadsMainDocument", Logger.LOG_LEVEL_FLOW);        	
    	}
    	else
    	{
    		gentleUpdateHeadsMainDocument();
    		Logger.sendMessage("RuleTranslator.updateHeadsMainDocument after gentleUpdateHeadsMainDocument", Logger.LOG_LEVEL_FLOW);        	
    	}
*/
    	Logger.sendMessage("RuleTranslator.updateHeadsMainDocument ended", Logger.LOG_LEVEL_FLOW);        	
	}
    
    private boolean roughUpdateHeadsMainDocument() throws IOException, JDOMException
    {
//    	String mainDocContent = getAxmlDocumentContent(this.rule.getHeadRelationAdornedName());

		Logger.sendMessage("RuleTranslator.roughUpdateHeadsMainDocument started", Logger.LOG_LEVEL_FLOW);        	
    	
        SAXBuilder builder = new SAXBuilder();

		Logger.sendMessage("RuleTranslator.roughUpdateHeadsMainDocument after new SAXBuilder", Logger.LOG_LEVEL_FLOW);        	
        
        String docName = outputDirName + File.separator + this.rule.getHeadRelationAdornedName() + ".xml";
        Document doc = builder.build(new File(docName));

		Logger.sendMessage("RuleTranslator.roughUpdateHeadsMainDocument after builder.build", Logger.LOG_LEVEL_FLOW);        	
        
        Element root = doc.getRootElement();

		Logger.sendMessage("RuleTranslator.roughUpdateHeadsMainDocument after doc.getRootElement", Logger.LOG_LEVEL_FLOW);        	
        
        // If there exists a data element, it needs to be located
        // after the service call elements. So remove it here...
//        Element dataElem = root.getChild("data");
//        if(dataElem != null)
//        	root.removeChild("data");
        
        if (root.getChildren("Rule").size() > 0)
        {
    		Logger.sendMessage("RuleTranslator.roughUpdateHeadsMainDocument ended (false)", Logger.LOG_LEVEL_FLOW);        	
        	return false;
        }
        
        roughAddRuleElement(root);

		Logger.sendMessage("RuleTranslator.roughUpdateHeadsMainDocument after roughAddRuleElement", Logger.LOG_LEVEL_FLOW);        	
        
//	    addRegisterServiceCall(root);
//	    addRuleServiceCall(root);
        
        // If the relation in the head of the rule is the leader,
        // add the <termination> element for the termination
        // detection algorithm.
//        if(this.rule.isHeadRelationLeader())
//        	addTerminationElement(root);
        
        // ... and add it here.
//        if(dataElem != null)
//        	root.addContent(dataElem);
        	
        // Update file.
        FileOutputStream out = new FileOutputStream(docName);
        XMLOutputter serializer = new XMLOutputter();
        serializer.output(doc, out);

        Logger.sendMessage("RuleTranslator.roughUpdateHeadsMainDocument after serializer.output", Logger.LOG_LEVEL_FLOW);        	

        out.flush();
        out.close();

        Logger.sendMessage("RuleTranslator.roughUpdateHeadsMainDocument after out.close", Logger.LOG_LEVEL_FLOW);        	
        
        this.axmlRepository.updateDocument(this.rule.getHeadRelationAdornedName());

		Logger.sendMessage("RuleTranslator.roughUpdateHeadsMainDocument ended (true)", Logger.LOG_LEVEL_FLOW);        	
        
        return true;
    }

    private void roughAddRuleElement(Element root)
    {
    	Element ruleElem = new Element("Rule");
    	
	    addRegisterServiceCall(ruleElem);
	    addRuleServiceCall(ruleElem);
	    
	    root.addContent(ruleElem);
    }
    
    private void gentleUpdateHeadsMainDocument() throws IOException, JDOMException
    {
		Logger.sendMessage("RuleTranslator.gentleUpdateHeadsMainDocument started", Logger.LOG_LEVEL_FLOW);        	
    	
        Element root = new Element("callAddRule");
        gentleAddRuleElement(root);
        Document outDoc = new Document(root);

		Logger.sendMessage("RuleTranslator.gentleUpdateHeadsMainDocument after new Document", Logger.LOG_LEVEL_FLOW);        	
        
        // Serialize it into a file.
        String docName = "callAddRule_" + this.rule.getHeadRelationAdornedName() + "_" + System.currentTimeMillis(); 
        FileOutputStream out = new FileOutputStream(outputDirName + File.separator + docName + ".xml");
        XMLOutputter serializer = new XMLOutputter();
        serializer.output(outDoc, out);
        out.flush();
        out.close();

		Logger.sendMessage("RuleTranslator.gentleUpdateHeadsMainDocument after out.close", Logger.LOG_LEVEL_FLOW);        	
        
        this.axmlRepository.addDocument(docName);

        Logger.sendMessage("RuleTranslator.gentleUpdateHeadsMainDocument ended", Logger.LOG_LEVEL_FLOW);        	
   }
    
    private void gentleAddRuleElement(Element root)
    {
    	// Create the addRule sc element.
    	String addRuleServiceName = "addRule_" + this.rule.getHeadRelationAdornedName();
    	Element addRuleScElement = Qsq2AxmlUtils.createScElement(
    			null, "once", addRuleServiceName, addRuleServiceName, 
    			this.localURL, null, null, null, null);
    	
    	// Add params to service call.
    	Element paramsElem = new Element("params", axmlNamespace);
    	Element ruleParamElem = new Element("param", axmlNamespace).setAttribute("name", "rule");
    	Element ruleParamVal = new Element("value", axmlNamespace);

    	// The parameter is the <Rule> element with its 2 service call
    	// children.
    	Element ruleElem = new Element("Rule");
	    addRegisterServiceCall(ruleElem);
	    addRuleServiceCall(ruleElem);
	    
	    ruleParamVal.addContent(ruleElem);
    	ruleParamElem.addContent(ruleParamVal);
    	paramsElem.addContent(ruleParamElem);
    	addRuleScElement.addContent(paramsElem);
	    root.addContent(addRuleScElement);
    }
    
    private void addRegisterServiceCall(Element ruleElem)
    {
        // Add sc element.
    	Element scElement = Qsq2AxmlUtils.createScElement(null, "once", "register", "RegisterService", 
    			this.localURL, null, null, null,
				/* followedBy */ "../axml:sc[@serviceNameSpace='ContRule_" + name + "']");
    	
    	// Add params to service call.
    	Element paramsElem = new Element("params", axmlNamespace);

    	Element serviceNameParamElem = new Element("param", axmlNamespace).setAttribute("name", "serviceName");
    	serviceNameParamElem.addContent(new Element("value", axmlNamespace).setText("Rule_" + name));
    	paramsElem.addContent(serviceNameParamElem);

    	Element clientIDParamElem = new Element("param", axmlNamespace).setAttribute("name", "clientID");
    	clientIDParamElem.addContent(new Element("xpath", axmlNamespace).setText("./@id"));
    	paramsElem.addContent(clientIDParamElem);

    	Element clientURLParamElem = new Element("param", axmlNamespace).setAttribute("name", "clientURL");
    	clientURLParamElem.addContent(new Element("value", axmlNamespace));
    	paramsElem.addContent(clientURLParamElem);

    	Element clientRelationNameParamElem = new Element("param", axmlNamespace).setAttribute("name", "clientRelationName");
    	clientRelationNameParamElem.addContent(new Element("value", axmlNamespace));
    	paramsElem.addContent(clientRelationNameParamElem);

    	Element servicePeerParamElem = new Element("param", axmlNamespace).setAttribute("name", "servicePeer");
    	servicePeerParamElem.addContent(new Element("value", axmlNamespace).setText(this.localPeerName));
    	paramsElem.addContent(servicePeerParamElem);

    	Element clientPeerParamElem = new Element("param", axmlNamespace).setAttribute("name", "clientPeer");
    	clientPeerParamElem.addContent(new Element("value", axmlNamespace).setText(this.localPeerName));
    	paramsElem.addContent(clientPeerParamElem);

    	scElement.addContent(paramsElem);	
    	
    	ruleElem.addContent(scElement);   	    	    	
    }
    
    private void addRuleServiceCall(Element ruleElem)
    {
        // Add sc element.
    	Element scElement = Qsq2AxmlUtils.createScElement(this.serviceCallFrequency, "lazy", "ContRule_" + name, "ContRule_" + name, 
    			this.localURL, null, /* subscription */ "true", /* useDiff */ "true", null); 
        
        // Add merge rule to avoid duplication of data.
//    	Element mergeElement = (new Element("merge-element-rule", axmlNamespace)).setAttribute("mode", "merge");
//    	Element keyElement = new Element("key", axmlNamespace);
//    	keyElement.addContent((new Element("selector", axmlNamespace)).setAttribute("xpath", "/" + this.head.getName() + "/data/tuples"));   	   	
//    	mergeElement.addContent(keyElement);	
//    	scElement.addContent(mergeElement);	

    	// Add params to service call.
    	Element paramsElem = new Element("params", axmlNamespace);

    	Element clientIDParamElem = new Element("param", axmlNamespace).setAttribute("name", "clientID");
    	clientIDParamElem.addContent(new Element("xpath", axmlNamespace).setText("../axml:sc[@serviceNameSpace='RegisterService']/@id"));
    	paramsElem.addContent(clientIDParamElem);

    	Element servicePeerParamElem = new Element("param", axmlNamespace).setAttribute("name", "servicePeer");
    	servicePeerParamElem.addContent(new Element("value", axmlNamespace).setText(this.localPeerName));
    	paramsElem.addContent(servicePeerParamElem);

    	scElement.addContent(paramsElem);	
    	
    	ruleElem.addContent(scElement);   	    	
    }

/*    
    private String getAxmlDocumentContent(String docName)
    {
		try
		{
			PeerAccessPointServiceLocator locator = new PeerAccessPointServiceLocator();
			URL serviceURL = new URL(localShortURL + "/services/PeerAccessPoint");
			localhost.axml.services.PeerAccessPoint.PeerAccessPoint service = locator.getPeerAccessPoint(serviceURL);
			return service.getDocumentContent(docName);
		}
		catch(Exception e)
		{
			Logger.sendMessage("ERROR: Failed to get content of AXML document " + docName + ".",
							   Logger.LOG_LEVEL_ERROR);
			return null;
		}    	
    }
*/
    
    private boolean isFirstRuleInHeadsMainDocument() throws IOException, JDOMException
    {
        SAXBuilder builder = new SAXBuilder();
        String docName = outputDirName + File.separator + this.rule.getHeadRelationAdornedName() + ".xml";
        Document doc = builder.build(new File(docName));
        Element root = doc.getRootElement();
        if (root.getChildren("Rule").size() == 0)
        	return true;
        
        return false;
    }
}
