package il.ac.tau.cs.dqsq.termdetection;

import java.io.File;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import il.ac.tau.cs.dqsq.util.dQSQUtils;
import il.ac.tau.cs.dqsq.termdetection.ActivityService;
import il.ac.tau.cs.dqsq.axmlservices.RegisterService;

/**
 * @author pettelno
 *
 * Provides web service to manage the activity level of nodes, as a part of
 * the termination detection algorithm. The activity level is written to
 * the activity file activity.xml which is located in the AXML repository.
 * 
 */
public class ActivityService
{
	private static final int OPERATION_INCREMENT = 0;
	private static final int OPERATION_DECREMENT = 1;
		
	/**
	 * This method is exposed as a web service. It increments the activity
	 * of a node by 1.
	 * 
	 * @param relationName The name of the node (relation) to increment the
	 * 		  activity level.
	 * @param cause A description of the cause of this action (for debug
	 * 		  purposes).
	 * @param peer The name of the peer on which the node (relation) is
	 * 		  hosted.
	 * 
	 */
	public void incrementActivity(String relationName, String cause, String peer)
	{
		updateActivity(relationName, OPERATION_INCREMENT, 1, cause, peer);
	}

	/**
	 * This method is exposed as a web service. It decrements the activity
	 * of a node by 1.
	 * 
	 * @param relationName The name of the node (relation) to decrement the
	 * 		  activity level.
	 * @param cause A description of the cause of this action (for debug
	 * 		  purposes).
	 * @param peer The name of the peer on which the node (relation) is
	 * 		  hosted.
	 * 
	 */
	public void decrementActivity(String relationName, String cause, String peer)
	{
		updateActivity(relationName, OPERATION_DECREMENT, 1, cause, peer);
	}

	/**
	 * This method is exposed as a web service. It increments the activity
	 * of a node in the head of a rule after receiving data from the last
	 * relation in the rule body. The activity level is increased by the
	 * number of passive clients of the relation, since there is potentially
	 * new data for these clients. 
	 * 
	 * @param relationName The name of the node (relation) to increment the
	 * 		  activity level.
	 * @param peer The name of the peer on which the node (relation) is
	 * 		  hosted.
	 * @param serviceName The name of the service that sent the data to the
	 * 		  relation.
	 * 
	 */
	public void incrementActivityAfterDataFromSecondService(String relationName, String peer, String serviceName)
	{	
		int numOfPassiveClients = 0; 
		try
		{
//			numOfPassiveClients = new RegisterService().getNumOfPassiveClients(relationName, peer);
			numOfPassiveClients = new RegisterService().getNumOfClients(relationName, peer);
			dQSQUtils.debug("DEBUG - " + relationName + " has " + numOfPassiveClients + " passive clients", "debug_Activity.txt");
		}
		catch(Exception ex)
		{
			dQSQUtils.debug("ERROR - ActivityService failed to get number of passive clients for relation " + relationName, "debug_Activity.txt");
			return;
		}
		
		// Increment activity by the number of passive clients i.e., clients
		// for which the "shouldCheckData" attribute is false. Do not increment
		// activity for clients with shouldCheckData=true, to avoid duplicity.
		// Such duplicity can occur, for example, when a client registers on
		// a service, and before the service checks for data for this client,
		// it receives new data from an external resource. In this case, we
		// shouldn't increment the activity value twice for this client.
		updateActivity(relationName, OPERATION_INCREMENT, numOfPassiveClients + 1, "rule 5, after receiving data from second relation " + serviceName, peer);
	}

/*	
	public void incrementActivityByNumOfClients(String relationName, String repositoryLocation)
	{
		String clientsFileName = repositoryLocation + File.separator + relation2Service(relationName) + "_clients.xml";
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		
		try
		{
			Document clientsDoc = dbf.newDocumentBuilder().parse(new File(clientsFileName));
			Element docElem = clientsDoc.getDocumentElement();
			NodeList clientElems = docElem.getChildNodes();
			int clientsNum = clientElems.getLength() - 1;
			
			// Get the number of clients for which the "shouldCheckData"
			// attribute is false. Do not increment activity for clients\
			// with shouldCheckData=true, to avoid duplicity. For example,
			
			// Increment actitivty by clientsNum + 1.
			updateActivity(relationName, repositoryLocation, OPERATION_INCREMENT, clientsNum + 1, "rule 5, after receiving data from second relation");
		}
		catch(Exception ex)
		{
			System.out.println("ERROR - Qsq2AxmlActivity.incrementActivityByNumOfClients failed");
		}
	}
*/

	public boolean isActive(String relationName, String peer)
	{
		// A node is active iff its activity level is greater than 0.
		
		String activityFileName = dQSQUtils.getRepositoryLocation() + File.separator + "activity.xml";	
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
	
		Document activityDoc = null;
		try
		{
			activityDoc = dbf.newDocumentBuilder().parse(new File(activityFileName));
		}
		catch(Exception ex)
		{
			dQSQUtils.debug("ERROR - ActivityService.isActive: Failed to parse activity file. System error: " + ex.getMessage(), "debug_Activity.txt");
			return true; // exception
		}

		Element activityLevelElement = getActivityLevelElement(activityDoc, relationName);
		if(activityLevelElement == null)
		{
			dQSQUtils.debug("ERROR - Failed to check activity - relation: " + relationName + ".", "debug_Activity.txt");
			return true; // exception
		}
		
		if(getElementIntValue(activityLevelElement) > 0)
			return true;
		
		return false;
	}

	/**
	 * This method is exposed as a web service. It checks whether the
	 * given peer is terminated.
	 * 
	 * @param peer The name of the peer.
	 * @return True if the peer is terminated, false otherwise.
	 */
	public boolean isTerminated(String peer)
	{
		// The peer is terminated iff the value of the "terminated" 
		// element in activity.xml is "y".
		
		String activityFileName = dQSQUtils.getRepositoryLocation() + File.separator + "activity.xml";
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
	
		Document activityDoc = null;
		try
		{
			activityDoc = dbf.newDocumentBuilder().parse(new File(activityFileName));
		}
		catch(Exception ex)
		{
			dQSQUtils.debug("ERROR - ActivityService.isTerminated: Failed to parse activity file. System error: " + ex.getMessage(), "debug_Activity.txt");
			return false; 
		}
	
		Element docElem = activityDoc.getDocumentElement();
		if(docElem.getAttribute("terminated").equals("y"))
			return true;
		
		return false;		
	}
	
	public void setTerminated(String peer)
	{
		String activityFileName = dQSQUtils.getRepositoryLocation() + File.separator + "activity.xml";
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
	
		Document activityDoc = null;
		try
		{
			activityDoc = dbf.newDocumentBuilder().parse(new File(activityFileName));
		}
		catch(Exception ex)
		{
			dQSQUtils.debug("ERROR - ActivityService.setTerminated: Failed to parse activity file. System error: " + ex.getMessage(), "debug_Activity.txt");
			return; // exception
		}
	
		Element docElem = activityDoc.getDocumentElement();
		docElem.setAttribute("terminated", "y");
		dQSQUtils.flushDocument(activityDoc, activityFileName);					
	}
	
	private Element getActivityLevelElement(Document activityDoc, String relationName)
	{				
		Element docElem = activityDoc.getDocumentElement();
		NodeList relationElems = docElem.getElementsByTagName("relation");
		int relationsNum = relationElems.getLength();		
		Element relationElement = null;
		
		for(int i = 0; i < relationsNum; i++)
		{
			Element curRelationElement = (Element)relationElems.item(i);
			String curRelationName = curRelationElement.getAttribute("name");
			if(curRelationName.equals(relationName))
			{
				relationElement = curRelationElement;
				break;
			}
		}
		
		if(relationElement == null)
			return null; // Relation not found in activity doc
		
		return (Element)relationElement.getElementsByTagName("activityLevel").item(0);
	}
	
	private void updateActivity(String relationName, int operation, int value, String cause, String peer)
	{
		synchronized(this.getClass())
		{
			dQSQUtils.debug("Updating activity - relation: " + relationName + " operation: " + operation + " value: " + value + "(" + cause + ")", "debug_Activity.txt");
			
			String activityFileName = dQSQUtils.getRepositoryLocation() + File.separator + "activity.xml";
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();

			Document activityDoc = null;
			try
			{
				activityDoc = dbf.newDocumentBuilder().parse(new File(activityFileName));
			}
			catch(Exception ex)
			{
				dQSQUtils.debug("ERROR - ActivityService.updateActivity: Failed to parse activity file. System error: " + ex.getMessage(), "debug_Activity.txt");
				return;
			}

			Element activityLevelElement = getActivityLevelElement(activityDoc, relationName);
			if(activityLevelElement == null)
			{
				dQSQUtils.debug("ERROR - Failed to update activity - relation: " + relationName + " operation: " + operation + " value: " + value, "debug_Activity.txt");
				return ; // error
			}

			switch(operation)
			{
				case OPERATION_INCREMENT:
					incrementElementValue(activityLevelElement, value);
					break;
				case OPERATION_DECREMENT:
					decrementElementValue(activityLevelElement, value);
					break;						
			}
				
			dQSQUtils.flushDocument(activityDoc, activityFileName);			
		}
	}
	
	private void incrementElementValue(Element elem, int value)
	{
		Node valueNode = elem.getFirstChild();
		int intVal = new Integer(valueNode.getNodeValue()).intValue();
		intVal += value;
		valueNode.setNodeValue(new String("" + intVal));
	}

	private void decrementElementValue(Element elem, int value)
	{
		Node valueNode = elem.getFirstChild();
		int intVal = new Integer(valueNode.getNodeValue()).intValue();
		intVal -= value;
		valueNode.setNodeValue(new String("" + intVal));
	}
	
	private int getElementIntValue(Element elem)
	{
		Node valueNode = elem.getFirstChild();
		int intVal = new Integer(valueNode.getNodeValue()).intValue();
		return intVal;
	}
	
	/**
	 * This method is exposed as a web service. It adds a new relation to
	 * the activity file while the AXML peer is running.
	 * 
	 * @param peerName The name of the peer on which the relation should be
	 *        added.
	 * @param relationName The name of the relation to add.
	 * 
	 */
	public void addRelation(String peerName, String relationName)
	{
		synchronized(this.getClass())
		{
			dQSQUtils.debug("Adding new relation: " + relationName, "debug_Activity.txt");
			
			String activityFileName = dQSQUtils.getRepositoryLocation() + File.separator + "activity.xml";
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
	
			Document activityDoc = null; 
			try
			{
				activityDoc = dbf.newDocumentBuilder().parse(new File(activityFileName));
			}
			catch(Exception ex)
			{
				dQSQUtils.debug("ERROR - ActivityService failed to parse activity file.", "debug_RegisterService.txt");
				return;
			}
				
			Element docElem = activityDoc.getDocumentElement();
							
			Element clientElem = activityDoc.createElement("relation");
			clientElem.setAttribute("name", relationName);
			Element activityLevelElem = activityDoc.createElement("activityLevel");
			Node activityLevelVal = activityDoc.createTextNode("0");
			activityLevelElem.appendChild(activityLevelVal);
			clientElem.appendChild(activityLevelElem);
			docElem.appendChild(clientElem);
			dQSQUtils.flushDocument(activityDoc, activityFileName);
		}		
	}
}
