package il.ac.tau.cs.dqsq.termdetection;

import java.net.URL;
import java.util.Map;
import java.util.HashMap;
import java.util.Collection;
import java.util.Iterator;
import javax.xml.parsers.DocumentBuilderFactory;
import localhost.axml.services.Datalog2QSQ.PeerServiceLocator;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import il.ac.tau.cs.dqsq.util.dQSQUtils;

/**
 * @author pettelno
 *
 * Manages the termination detection (TD) algorithm for an AXML peer, to
 * detect when the dQSQ computation is finished.
 * This class exposes the web service isFinished which is called
 * periodically by the leader node. Such a call initilates a TD cycle.
 * In our implementation a node in the TD algorithm is a dQSQ relation.
 * 
 */
public class IdlenessManager
{
	private String localPeerName = null;
	private int lastCycleID = 0;
	private boolean servicesTerminated = false;
//	private boolean initialized = false;

	// A Map of LocalNodeIdlenessManager objects, which contains all the
	// nodes (relations) on this peer.
	private Map relations = new HashMap(); 
	
	public IdlenessManager()
	{
		dQSQUtils.debug("DEBUG - IdlenessManager created.", "debug_Idleness.txt");
	}

	/**
	 * This method is exposed as a web service which is called periodically
	 * by the leader node to check if the dQSQ computation is finished.  
     *
	 * @param nodeName The name of the calling node.
	 * @param servicePeer The peer that hosts the calling node.
	 * @return An Element with either a negative or a positive reply, which
	 *         is inserted directly to the calling AXML document.
	 * 
	 */
	/* synchronized */ public Element isFinished(String nodeName, String servicePeer)
	{
		Element replyElem = null;
		try
		{
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			dbf.setNamespaceAware(true);
			Document doc = dbf.newDocumentBuilder().newDocument();
			replyElem = doc.createElement("reply");

			lastCycleID++;
			dQSQUtils.debug("----------------------------------------------", "debug_Idleness.txt");
			dQSQUtils.debug("DEBUG - isFinished called for node " + nodeName + ". Cycle ID: " + lastCycleID, "debug_Idleness.txt");
			if(isIdle(nodeName, lastCycleID, servicePeer) == Constants.STATUS_IDLE)
			{
				dQSQUtils.debug("DEBUG - isFinished returned true for node "+ nodeName, "debug_Idleness.txt");
				
				setPositiveReply(replyElem, doc);
				terminateServices(servicePeer);
				informPeerQueryComputationFinished();				
			}
			else
			{
				dQSQUtils.debug("DEBUG - isFinished returned false for node "+ nodeName, "debug_Idleness.txt");
				setNegativeReply(replyElem, doc);
			}
		}
		catch(Exception ex)
		{
			dQSQUtils.debug("ERROR - isFinished failed for node " + nodeName + ". Cycle ID: " + lastCycleID, "debug_Idleness.txt");			
		}
			
		return replyElem;
	}
	
/*
	synchronized public Element isFinished(String nodeName, String servicePeer)
	{
		Element replyElem = null;
		try
		{
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			dbf.setNamespaceAware(true);
			Document doc = dbf.newDocumentBuilder().newDocument();
			replyElem = doc.createElement("reply");

			lastCycleID++;
			dQSQUtils.debug("DEBUG - isFinished called for node " + nodeName + ". Cycle ID: " + lastCycleID, "debug_Idleness.txt");
			if(isIdle(nodeName, lastCycleID, servicePeer))
			{
				dQSQUtils.debug("DEBUG - isFinished returned true for node "+ nodeName, "debug_Idleness.txt");
				
				setPositiveReply(replyElem, doc);
				terminateServices(servicePeer);
			}
			else
			{
				dQSQUtils.debug("DEBUG - isFinished returned false for node "+ nodeName, "debug_Idleness.txt");
				setNegativeReply(replyElem, doc);
			}
		}
		catch(Exception ex)
		{
			dQSQUtils.debug("ERROR - isFinished failed for node " + nodeName + ". Cycle ID: " + lastCycleID, "debug_Idleness.txt");			
		}
			
		return replyElem;
	}
*/
	
	private void setPositiveReply(Element replyElem, Document doc)
	{
		// Build the "finished" element and add it to the reply.
		Element finishedElem = doc.createElement("finished");
		Node valueNode = doc.createTextNode("y");
		finishedElem.appendChild(valueNode);
		replyElem.appendChild(finishedElem);
		
		// Build the service call element to stop the continuous
		// service "ContIsFinished" and add it to the reply.
		Element serviceElem = doc.createElementNS("http://www-rocq.inria.fr/verso/AXML", "sc");
		serviceElem.setAttributeNS("http://www.w3.org/2000/xmlns/",
								   "xmlns:axml", 
								   "http://www-rocq.inria.fr/verso/AXML");
		serviceElem.setAttribute("frequency", "once");
		serviceElem.setAttribute("methodName", "ContIsFinished");
		serviceElem.setAttribute("serviceNameSpace", "ContIsFinished");
		serviceElem.setAttribute("serviceURL", dQSQUtils.getPeerURL() + "/servlet/AxisServlet");
		serviceElem.setAttribute("subscription", "true");
		serviceElem.setAttribute("destroy", "true");
		replyElem.appendChild(serviceElem);
	}
	
	private void setNegativeReply(Element replyElem, Document doc)
	{
		// Build the "finished" element and add it to the reply.
		Element finishedElem = doc.createElement("finished");
		Node valueNode = doc.createTextNode("n");
		finishedElem.appendChild(valueNode);
		replyElem.appendChild(finishedElem);
/*		
 		This is the way to add a service call - keep this code!!!
 		---------------------------------------------------------
		// Build the service call element and add it to the reply.
		Element serviceElem = doc.createElementNS("http://www-rocq.inria.fr/verso/AXML", "sc");
		serviceElem.setAttributeNS("http://www.w3.org/2000/xmlns/",
								   "xmlns:axml", 
								   "http://www-rocq.inria.fr/verso/AXML");
		serviceElem.setAttribute("frequency", "once");
		serviceElem.setAttribute("methodName", "isFinished");
		serviceElem.setAttribute("serviceNameSpace", "IdlenessManager");
		serviceElem.setAttribute("serviceURL", "http://localhost:8080/axml/servlet/AxisServlet");
		replyElem.appendChild(serviceElem);
*/		
	}
	
	/**
	 * This method is exposed as a web service. It is called by 
	 * RemoteRelationIdlenessManager to delegate the TD cycle to a node 
	 * which is located on a remote peer. 
	 * 
	 * @param nodeName The name of the node to which the TD cycle is delegated. 
	 * @param cycleID The TD cycle ID.
	 * @param peer The peer that hosts the node to which the TD cycle is
	 *        delegated. 
	 */
	public int isIdle(String nodeName, int cycleID, String peer)
	{
/*		
		if(!initialized)
		{
			readNeighborsFile(peer);
			initialized = true;
		}
*/
		
		LocalRelationIdlenessManager rel = (LocalRelationIdlenessManager)relations.get(nodeName);
		if(rel != null)
			return rel.isIdle(cycleID, peer);
		
		dQSQUtils.debug("ERROR - IdlenessManager failed to find relation object for " + nodeName, "debug_Idleness.txt");
		return Constants.STATUS_ACTIVE;
	}

	/**
	 * This method is exposed as a web service. Whenever a node receives a
	 * message as a part of the dQSQ computation process, it calls this
	 * service.
	 * 
	 * @param relationName The name of the node that received a message.
	 * 
	 */
	public void onRecvMsg(String relationName)
	{
		LocalRelationIdlenessManager rel = (LocalRelationIdlenessManager)relations.get(relationName);
		if(rel != null)
			rel.onRecvMsg();
		else
			dQSQUtils.debug("ERROR - IdlenessManager failed to find relation object for " + relationName, "debug_Idleness.txt");
	}

	/**
	 * This method is exposed as a web service. It is used to mark that
	 * the peer is terminated in activity.xml after termination is detected.
	 * 
	 * @param servicePeer The name of the peer to mark as terminated.
	 * 
	 */
	/* synchronized */ public void terminateServices(String servicePeer)
	{
		if(!servicesTerminated)
		{
			servicesTerminated = true;
			new ActivityService().setTerminated(servicePeer);
			
			// Terminate services for all remote neighbors recursively.
			terminateRemoteNeighborServices();
			
			clear();			
		}
	}

/*	
	private void readNeighborsFile(String peer)
	{
		String neighborsFileName = null; 
		try
		{
			neighborsFileName = dQSQUtils.getRepositoryLocation(peer) + File.separator + "neighbors.xml";
		}
		catch(Exception ex)
		{
			dQSQUtils.debug("ERROR - IdlenessManager failed to get repository location", "debug_Idleness.txt");
			return;
		}
		
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		
		Document neighborsDoc = null; 
		try
		{
			neighborsDoc = dbf.newDocumentBuilder().parse(new File(neighborsFileName));
		}
		catch(Exception ex)
		{
			dQSQUtils.debug("ERROR - IdlenessManager failed to parse neighbors file", "debug_Idleness.txt");
			return;
		}
		
		Element docElem = neighborsDoc.getDocumentElement();
		localPeerName = docElem.getAttribute("name");

		initRelations(docElem);
	}
*/

/*	
	private void initRelations(Element docElem)
	{
		// 1. Create relation objects and put them in the map.
		NodeList relationElems = docElem.getElementsByTagName("relation");
		int relationsNum = relationElems.getLength();
		for(int i = 0; i < relationsNum; i++)
		{
			Element curRelationElem = (Element)relationElems.item(i);
			String curRelationName = curRelationElem.getAttribute("name");
			
			// Add the relation to the map if not already exists.
			if(relations.get(curRelationName) == null)
				relations.put(curRelationName, new LocalRelationIdlenessManager(curRelationName, localPeerName));
		}

		// 2. Set relations neighbors.
		// Need to go over the relation nodes again, because setting
		// of neighbors can be done only after all the nodes objects
		// are created.
		for(int i = 0; i < relationsNum; i++)
		{
			Element curRelationElem = (Element)relationElems.item(i);
			String curRelationName = curRelationElem.getAttribute("name");
			LocalRelationIdlenessManager curRelation = (LocalRelationIdlenessManager)relations.get(curRelationName);
			curRelation.setNeighbors(curRelationElem, relations);
		}
	}
*/
	
	private void terminateRemoteNeighborServices()
	{
		Collection rels = relations.values();
		Iterator iter = rels.iterator();
		while(iter.hasNext())
		{
			LocalRelationIdlenessManager currentRelation = (LocalRelationIdlenessManager)iter.next();
			currentRelation.terminateRemoteNeighborServices();
		}
	}
	
	/**
	 * This method is exposed as a web service. Whenever a new node is
	 * added dynamically (while the AXML peer is already running) to the
	 * neighbors file (neighbors.xml), this service is called, and the
	 * neighbors map is refreshed. 
	 * 
	 * @param peer The name of the peer whose neighbors map should be refreshed.
	 * 
	 */
	/*
	synchronized public void refreshNeighborsMap(String peer)
	{
		dQSQUtils.debug("DEBUG - IdlenessManager refreshing neighbors map for peer " + peer, "debug_Idleness.txt");
		readNeighborsFile(peer);
	}
	*/
	
	
	/* synchronized */ public void clear()
	{
		dQSQUtils.debug("DEBUG - IdlenessManager is clearing.", "debug_Idleness.txt");

		this.lastCycleID = 0;
		this.servicesTerminated = false;
//		this.initialized = false;
		this.relations = new HashMap(); 
	}
	
	/**
	 * This method is exposed as a web service. It adds a new relation
	 * to the map of relation nodes.
	 * 
	 * @param relationName The name of the relation to add.
	 * 
	 */
	public void addRelation(String relationName)
	{
		LocalRelationIdlenessManager rel = (LocalRelationIdlenessManager)this.relations.get(relationName);
		if(rel == null)
		{
			this.relations.put(
					relationName,
					new LocalRelationIdlenessManager(relationName, localPeerName));
	
			dQSQUtils.debug("DEBUG - IdlenessManager added relation " + relationName + " to peer " + localPeerName, "debug_Idleness.txt");
		}
	}
	
	/**
	 * This method is exposed as a web service. It adds a neighbor 
	 * relation to an already existing relation.
	 * 
	 * @param relationName The name of the relation to which the 
	 *        neighbor is added.
	 * @param neighborName The name of the neighbor to add.
	 * @param neighborPeerName The peer name of the neighbor.
	 * @param neighborURL The URL of the neighbor.
	 * 
	 */
	public void addNeighbor(String relationName, String neighborName,
							String neighborPeerName, String neighborURL)
	{
		LocalRelationIdlenessManager relation =
			(LocalRelationIdlenessManager)this.relations.get(relationName);
		if(relation == null)
			dQSQUtils.debug("ERROR - IdlenessManager failed to add neighbor " + 
							neighborName + " to relation " + relationName, "debug_Idleness.txt");
		
		relation.addNeighbor(neighborName, neighborPeerName,
							 neighborURL, this.relations);
		
		dQSQUtils.debug("DEBUG - IdlenessManager added neighbor " + neighborName + " to relation " + relationName + " on peer " + localPeerName, "debug_Idleness.txt");		
	}

	/**
	 * This method is exposed as a web service.
	 * 
	 * @param peerName
	 * 
	 */
	public void init(String peerName)
	{
		this.localPeerName = peerName;
	}

/*	
	private void sendTerminationMessageToPeerLog()
	{
    	PeerServiceLocator locator = new PeerServiceLocator();
		try
		{
			URL serviceURL = new URL(dQSQUtils.getPeerURL() + "/services/Datalog2QSQ");
			localhost.axml.services.Datalog2QSQ.Peer remotePeer =
				locator.getDatalog2QSQ(serviceURL);
			remotePeer.sendLogMessage("Query computation is finished.", Logger.LOG_LEVEL_DEBUG);
		}
		catch(Exception e)
		{
		}    			
	}
*/

	private void informPeerQueryComputationFinished()
	{
    	PeerServiceLocator locator = new PeerServiceLocator();
		try
		{
			URL serviceURL = new URL(dQSQUtils.getPeerURL() + "/services/Datalog2QSQ");
			localhost.axml.services.Datalog2QSQ.Peer remotePeer =
				locator.getDatalog2QSQ(serviceURL);
			remotePeer.queryComputationFinished();
		}
		catch(Exception e)
		{
		}    			
	}	
}
