package il.ac.tau.cs.dqsq.termdetection;

import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import java.util.Map;
import java.util.HashMap;
import java.util.Collection;
import java.util.Iterator;
import il.ac.tau.cs.dqsq.util.dQSQUtils;

/**
 * @author pettelno
 *
 * A node in the termination detection (TD) algorithm, which is local
 * (representing a node that physically located on this peer).
 * In our implementation a node in the TD algorithm is a dQSQ relation. 
 * 
 */
class LocalRelationIdlenessManager extends RelationIdlenessManager
{
	private String localPeerName = null;
	private int idleness = 0;
	private int lastCycleID = 0;
	
	// A Map of NodeIdlenessManager objects, some are LocalNodeIdlenessManager
	// and some are RemoteNodeIdlenessManager.
	private Map neighborRelations = new HashMap(); 
	
	public LocalRelationIdlenessManager(String relationName, String localPeerName)
	{
		this.relationName = relationName;
		this.localPeerName = localPeerName;
	}
	
	public int isIdle(int cycleID, String peer)
	{
		// isIdle returns one of three values:
		// * STATUS_IDLE if the node is idle.
		// * STATUS_ACTIVE if the node is active.
		// * STATUS_PRE_IDLE if the node and all its neighbors are idle,
		//   but the idleness value of the node or one of the neighbors is
		//   smaller than 2. According to the termination detection
		//   algorithm, the computation is terminated if all the nodes are
		//   idle for 2 consecutive cycles.
		//
		// The isIdle algorithm:
		// 1. If I already participated in this TD cycle - return STATUS_IDLE.
		// 2. Else if I am active - return STATUS_ACTIVE.
		// 3. Else if at least one of my neighbors is active - return STATUS_ACTIVE.
		// 4. Else if I am pre-idle - return STATUS_PRE_IDLE.  
		// 5. Else if at least one of my neighbors is pre-idle - return STATUS_PRE_IDLE.
		// 6. Else I and all my neighbors are idle - return STATUS_IDLE.
		//
		// The algorithm has the following optimization: If one of my 
		// neighbors is pre-idle, I don't return immediately. Instead,
		// I continue to check at all the other neighbors. This way we reduce
		// the time we detect termination after the activity of all nodes
		// become 0. In this case all nodes will increase their idleness
		// value to 1 IN THE SAME cycle, instead of increasing the idleness
		// value only for a single mode per cycle.
		
		dQSQUtils.debug("DEBUG - LocalRelationIdlenessManager.isIdle: Checking idleness for relation " + relationName, "debug_Idleness.txt");					

		if(cycleID <= lastCycleID)
		{
			dQSQUtils.debug("DEBUG - LocalRelationIdlenessManager.isIdle: Relation " + relationName + " already participated in this cycle - returning true", "debug_Idleness.txt");					
			return Constants.STATUS_IDLE; // already participated in this cycle
		}
		
		lastCycleID = cycleID;
		
		if(new ActivityService().isActive(relationName, peer))
		{
			dQSQUtils.debug("DEBUG - LocalRelationIdlenessManager.isIdle: Relation " + relationName + " is active", "debug_Idleness.txt");					
			return Constants.STATUS_ACTIVE;
		}
		
		boolean preIdleNeighborExists = false;
		Collection neighbors = neighborRelations.values();
		Iterator neighborsIter = neighbors.iterator();
		while(neighborsIter.hasNext())
		{
			RelationIdlenessManager relation = (RelationIdlenessManager)neighborsIter.next();
			int neighborIdleStatus = relation.isIdle(cycleID, peer);
			if(neighborIdleStatus == Constants.STATUS_ACTIVE)
			{
				// One of my neighbors is active.
				dQSQUtils.debug("DEBUG - LocalRelationIdlenessManager.isIdle: Neighbor relation " + relation.relationName + " is active", "debug_Idleness.txt");
				return Constants.STATUS_ACTIVE;
			}
			if(neighborIdleStatus == Constants.STATUS_PRE_IDLE)
				preIdleNeighborExists = true;
		}

		// All my neighbors are either idle or pre-idle.
		idleness++;
		if(idleness < 2)
		{
			// I'm pre-idle.
			dQSQUtils.debug("DEBUG - LocalRelationIdlenessManager.isIdle: Relation " + relationName + " is pre-idle", "debug_Idleness.txt");
			return Constants.STATUS_PRE_IDLE; 
		}
		
		// I'm idle. If all my neighbors are also idle - return STATUS_IDLE,
		// else - return STATUS_PRE_IDLE.
		if(preIdleNeighborExists)
		{
			// Not all neighbors are idle. There is at least one pre-idle
			// neighbor.
			dQSQUtils.debug("DEBUG - LocalRelationIdlenessManager.isIdle: Relation " + relationName + " is idle but one of its neighbors is pre-idle", "debug_Idleness.txt");
			return Constants.STATUS_PRE_IDLE;
		}
		
		// All my neighbors are also idle.
		dQSQUtils.debug("DEBUG - LocalRelationIdlenessManager.isIdle: Relation " + relationName + " and all its neighbors are idle", "debug_Idleness.txt");
		return Constants.STATUS_IDLE;
	}
	
	public void onRecvMsg()
	{
		idleness = 0;

		dQSQUtils.debug("DEBUG - Relation " + relationName + " received a message.", "debug_Idleness.txt");					
	}
	
	public void setNeighbors(Element curRelationElem, Map localRelations)
	{
		NodeList neighborElems = curRelationElem.getElementsByTagName("neighbor");
		int neighborsNum = neighborElems.getLength();
		RelationIdlenessManager rel = null; 
		for(int i = 0; i < neighborsNum; i++)
		{
			Element curNeighborElem = (Element)neighborElems.item(i);
			String curNeighborName = curNeighborElem.getAttribute("name");

			// Add the neighbot to the list if not already exists.
			if(neighborRelations.get(curNeighborName) == null)
			{
				String curNeighborPeer = curNeighborElem.getAttribute("peer");
				if(curNeighborPeer.equals(localPeerName))
				{
					// It's a local relation - take it from the local relations map.
					rel = (RelationIdlenessManager)localRelations.get(curNeighborName);
				}
				else
				{
					// It's a remote relation.
					String curNeighborURL = curNeighborElem.getAttribute("url");
					rel = new RemoteRelationIdlenessManager(curNeighborName, curNeighborURL, curNeighborPeer);
				}
	
				neighborRelations.put(curNeighborName, rel);
			}
		}
	}
	
	public void terminateRemoteNeighborServices()
	{
		Collection neighbors = neighborRelations.values();
		Iterator neighborsIter = neighbors.iterator();
		while(neighborsIter.hasNext())
		{
			RelationIdlenessManager relation = (RelationIdlenessManager)neighborsIter.next();
			relation.terminateServices();
		}		
	}
	
	void addNeighbor(String neighborName, String neighborPeerName,
					 String neighborURL, Map localRelations)
	{
		RelationIdlenessManager rel = null; 
		
		if(neighborRelations.get(neighborName) == null)
		{
			if(neighborPeerName.equals(localPeerName))
			{
				// It's a local relation - take it from the local relations map.
				rel = (RelationIdlenessManager)localRelations.get(neighborName);
			}
			else
			{
				// It's a remote relation.
				rel = new RemoteRelationIdlenessManager(neighborName, neighborURL, neighborPeerName);
			}

			neighborRelations.put(neighborName, rel);
		}		
	}
}
